# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Budget optimization converters.

This module defines various classes that convert `BudgetOptimizationResult`s
into flat dataframes.
"""

import abc
from collections.abc import Iterator, Sequence

from meridian import constants as c
from mmm.v1.marketing.optimization import budget_optimization_pb2 as budget_pb
from mmm.v1.marketing.optimization import constraints_pb2 as constraints_pb
from scenarioplanner.converters import mmm
from scenarioplanner.converters.dataframe import common
from scenarioplanner.converters.dataframe import constants as dc
from scenarioplanner.converters.dataframe import converter
import pandas as pd


__all__ = [
    "NamedOptimizationGridConverter",
    "BudgetOptimizationSpecsConverter",
    "BudgetOptimizationResultsConverter",
    "BudgetOptimizationResponseCurvesConverter",
]


class _BudgetOptimizationConverter(converter.Converter, abc.ABC):
  """An abstract class for dealing with `BudgetOptimizationResult`s."""

  def __call__(self) -> Iterator[tuple[str, pd.DataFrame]]:
    results = self._mmm.budget_optimization_results
    if not results:
      return

    # Validate group IDs.
    group_ids = [result.group_id for result in results if result.group_id]
    if len(set(group_ids)) != len(group_ids):
      raise ValueError(
          "Specified group_id must be unique or unset among the given group of"
          " results."
      )

    yield from self._handle_budget_optimization_results(
        self._mmm.budget_optimization_results
    )

  def _handle_budget_optimization_results(
      self, results: Sequence[mmm.BudgetOptimizationResult]
  ) -> Iterator[tuple[str, pd.DataFrame]]:
    raise NotImplementedError()


class NamedOptimizationGridConverter(_BudgetOptimizationConverter):
  """Outputs named tables for budget optimization grids.

  When called, this converter returns a data frame with the columns:

  *   "Group ID"
      A UUID generated for each named incremental outcome grid.
  *   "Channel"
  *   "Spend"
  *   "Incremental Outcome"
  For each named budget optimization result in the MMM output proto.
  """

  def _handle_budget_optimization_results(
      self, results: Sequence[mmm.BudgetOptimizationResult]
  ) -> Iterator[tuple[str, pd.DataFrame]]:
    for budget_opt_result in results:
      # There should be one unique ID for each result.
      group_id = (
          str(budget_opt_result.group_id) if budget_opt_result.group_id else ""
      )
      grid = budget_opt_result.incremental_outcome_grid

      # Each grid yields its own data frame table.
      optimization_grid_data = []
      for channel, cells in grid.channel_spend_grids.items():
        for spend, incremental_outcome in cells:
          optimization_grid_data.append([
              group_id,
              channel,
              spend,
              incremental_outcome,
          ])

      yield (
          common.create_grid_sheet_name(
              dc.OPTIMIZATION_GRID_NAME_PREFIX, grid.name
          ),
          pd.DataFrame(
              optimization_grid_data,
              columns=[
                  dc.OPTIMIZATION_GROUP_ID_COLUMN,
                  dc.OPTIMIZATION_CHANNEL_COLUMN,
                  dc.OPTIMIZATION_GRID_SPEND_COLUMN,
                  dc.OPTIMIZATION_GRID_INCREMENTAL_OUTCOME_COLUMN,
              ],
          ),
      )


class BudgetOptimizationSpecsConverter(_BudgetOptimizationConverter):
  """Outputs a table of budget optimization specs.

  When called, this converter returns a data frame with the columns:

  *   "Group ID"
      A UUID generated for an incremental outcome grid present in the output.
  *   "Date Interval Start"
  *   "Date Interval End"
  *   "Analysis Period"
  *   "Objective"
  *   "Scenario Type"
  *   "Initial Channel Spend"
  *   "Target Metric Constraint"
      None if scenario type is "Fixed"
  *   "Target Metric Value"
      None if scenario type is "Fixed"
  *   "Channel"
  *   "Channel Spend Min"
  *   "Channel Spend Max"
  """

  def _handle_budget_optimization_results(
      self, results: Sequence[mmm.BudgetOptimizationResult]
  ) -> Iterator[tuple[str, pd.DataFrame]]:
    spec_data = []
    for budget_opt_result in results:
      # There should be one unique ID for each result.
      group_id = (
          str(budget_opt_result.group_id) if budget_opt_result.group_id else ""
      )
      spec = budget_opt_result.spec

      objective = common.map_target_metric_str(spec.objective)
      # These are the start and end dates for the requested budget optimization
      # in this spec.
      date_interval_start, date_interval_end = (
          d.strftime(c.DATE_FORMAT) for d in spec.date_interval.date_interval
      )
      budget_date_interval = (date_interval_start, date_interval_end)

      # aka historical spend from marketing data in the model kernel
      initial_channel_spends = self._mmm.marketing_data.all_channel_spends(
          budget_date_interval
      )

      scenario = (
          dc.OPTIMIZATION_SPEC_SCENARIO_FIXED
          if spec.is_fixed_scenario
          else dc.OPTIMIZATION_SPEC_SCENARIO_FLEXIBLE
      )

      if spec.is_fixed_scenario:
        target_metric_constraint = None
        target_metric_value = None
      else:
        flexible_scenario = (
            spec.budget_optimization_spec_proto.flexible_budget_scenario
        )
        # Meridian flexible budget spec only has one target metric constraint.
        target_metric_constraint_pb = (
            flexible_scenario.target_metric_constraints[0]
        )
        target_metric_constraint = common.map_target_metric_str(
            target_metric_constraint_pb.target_metric
        )
        target_metric_value = target_metric_constraint_pb.target_value

      # When the constraint of a channel is not specified, that channel will
      # have a constraint of `[0, max_budget]` which is equivalent to no
      # constraint.
      #
      # Here, `max_budget` is the total budget for a fixed scenario spec, or the
      # max budget upper bound for a flexible scenario spec.
      #
      # NOTE: This assumption must be in line with what the budget optimization
      # processor does with an empty channel constraints list.
      channel_constraints = spec.channel_constraints
      if not channel_constraints:
        # Implicit channel constraints; synthesize them first before proceeding.
        channel_constraints = [
            budget_pb.ChannelConstraint(
                channel_name=channel_name,
                budget_constraint=constraints_pb.BudgetConstraint(
                    min_budget=0.0,
                    max_budget=spec.max_budget,
                ),
            )
            for channel_name in self._mmm.marketing_data.media_channels
        ]

      for channel_constraint in channel_constraints:
        spec_data.append([
            group_id,
            date_interval_start,
            date_interval_end,
            spec.date_interval_tag,
            objective,
            scenario,
            initial_channel_spends.get(channel_constraint.channel_name, 0.0),
            target_metric_constraint,
            target_metric_value,
            channel_constraint.channel_name,
            channel_constraint.budget_constraint.min_budget,
            channel_constraint.budget_constraint.max_budget,
        ])

    yield (
        dc.OPTIMIZATION_SPECS,
        pd.DataFrame(
            spec_data,
            columns=[
                dc.OPTIMIZATION_GROUP_ID_COLUMN,
                dc.OPTIMIZATION_SPEC_DATE_INTERVAL_START_COLUMN,
                dc.OPTIMIZATION_SPEC_DATE_INTERVAL_END_COLUMN,
                dc.ANALYSIS_PERIOD_COLUMN,
                dc.OPTIMIZATION_SPEC_OBJECTIVE_COLUMN,
                dc.OPTIMIZATION_SPEC_SCENARIO_TYPE_COLUMN,
                dc.OPTIMIZATION_SPEC_INITIAL_CHANNEL_SPEND_COLUMN,
                dc.OPTIMIZATION_SPEC_TARGET_METRIC_CONSTRAINT_COLUMN,
                dc.OPTIMIZATION_SPEC_TARGET_METRIC_VALUE_COLUMN,
                dc.OPTIMIZATION_CHANNEL_COLUMN,
                dc.OPTIMIZATION_SPEC_CHANNEL_SPEND_MIN_COLUMN,
                dc.OPTIMIZATION_SPEC_CHANNEL_SPEND_MAX_COLUMN,
            ],
        ),
    )


class BudgetOptimizationResultsConverter(_BudgetOptimizationConverter):
  """Outputs a table of budget optimization results objectives.

  When called, this converter returns a data frame with the columns:

  *   "Group ID"
      A UUID generated for a budget optimization result present in the output.
  *   "Channel"
  *   "Is Revenue KPI"
      Whether the KPI is revenue or not.
  *   "Optimal Spend"
  *   "Optimal Spend Share"
  *   "Optimal Impression Effectiveness"
  *   "Optimal ROI"
  *   "Optimal mROI"
  *   "Optimal CPC"
  """

  def _handle_budget_optimization_results(
      self, results: Sequence[mmm.BudgetOptimizationResult]
  ) -> Iterator[tuple[str, pd.DataFrame]]:
    data = []

    for budget_opt_result in results:
      group_id = (
          str(budget_opt_result.group_id) if budget_opt_result.group_id else ""
      )
      marketing_analysis = budget_opt_result.optimized_marketing_analysis

      media_channel_analyses = marketing_analysis.channel_mapped_media_analyses
      for channel, media_analysis in media_channel_analyses.items():
        # Skip "All Channels" pseudo-channel.
        if channel == c.ALL_CHANNELS:
          continue

        spend = media_analysis.spend_info_pb.spend
        spend_share = media_analysis.spend_info_pb.spend_share

        revenue_outcome = media_analysis.maybe_revenue_outcome
        nonrevenue_outcome = media_analysis.maybe_non_revenue_outcome

        # pylint: disable=cell-var-from-loop
        def _append_outcome_data(
            outcome: mmm.Outcome | None,
            is_revenue_kpi: bool,
        ) -> None:
          if outcome is None:
            return
          effectiveness = outcome.effectiveness_pb.value.value
          roi = outcome.roi_pb.value
          mroi = outcome.marginal_roi_pb.value
          cpc = outcome.cost_per_contribution_pb.value
          data.append([
              group_id,
              channel,
              is_revenue_kpi,
              spend,
              spend_share,
              effectiveness,
              roi,
              mroi,
              cpc,
          ])

        _append_outcome_data(revenue_outcome, True)
        _append_outcome_data(nonrevenue_outcome, False)
        # pylint: enable=cell-var-from-loop

    yield (
        dc.OPTIMIZATION_RESULTS,
        pd.DataFrame(
            data,
            columns=[
                dc.OPTIMIZATION_GROUP_ID_COLUMN,
                dc.OPTIMIZATION_CHANNEL_COLUMN,
                dc.OPTIMIZATION_RESULT_IS_REVENUE_KPI_COLUMN,
                dc.OPTIMIZATION_RESULT_SPEND_COLUMN,
                dc.OPTIMIZATION_RESULT_SPEND_SHARE_COLUMN,
                dc.OPTIMIZATION_RESULT_EFFECTIVENESS_COLUMN,
                dc.OPTIMIZATION_RESULT_ROI_COLUMN,
                dc.OPTIMIZATION_RESULT_MROI_COLUMN,
                dc.OPTIMIZATION_RESULT_CPC_COLUMN,
            ],
        ),
    )


class BudgetOptimizationResponseCurvesConverter(_BudgetOptimizationConverter):
  """Outputs a table of budget optimization response curves.

  When called, this converter returns a data frame with the columns:

  *   "Group ID"
      A UUID generated for a budget optimization result present in the output.
  *   "Channel"
  *   "Spend"
  *   "Incremental Outcome"
  """

  def _handle_budget_optimization_results(
      self, results: Sequence[mmm.BudgetOptimizationResult]
  ) -> Iterator[tuple[str, pd.DataFrame]]:
    response_curve_data = []
    for budget_opt_result in results:
      group_id = (
          str(budget_opt_result.group_id) if budget_opt_result.group_id else ""
      )
      curves = budget_opt_result.response_curves
      for curve in curves:
        for spend, incremental_outcome in curve.response_points:
          response_curve_data.append([
              group_id,
              curve.channel_name,
              spend,
              incremental_outcome,
          ])

    yield (
        dc.OPTIMIZATION_RESPONSE_CURVES,
        pd.DataFrame(
            response_curve_data,
            columns=[
                dc.OPTIMIZATION_GROUP_ID_COLUMN,
                dc.OPTIMIZATION_CHANNEL_COLUMN,
                dc.OPTIMIZATION_GRID_SPEND_COLUMN,
                dc.OPTIMIZATION_GRID_INCREMENTAL_OUTCOME_COLUMN,
            ],
        ),
    )


CONVERTERS = [
    NamedOptimizationGridConverter,
    BudgetOptimizationSpecsConverter,
    BudgetOptimizationResultsConverter,
    BudgetOptimizationResponseCurvesConverter,
]



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Serialization and deserialization of `InputData` for Meridian models."""

from collections.abc import Mapping
import dataclasses
import datetime as dt
import functools
import itertools
from typing import Sequence

from meridian import constants as c
from meridian.data import input_data as meridian_input_data
from mmm.v1.common import date_interval_pb2
from mmm.v1.marketing import marketing_data_pb2 as marketing_pb
from schema.serde import constants as sc
from schema.serde import serde
from schema.utils import time_record
import numpy as np
import xarray as xr

from google.type import date_pb2

# Mapping from DataArray names to coordinate names
_COORD_NAME_MAP = {
    c.MEDIA: c.MEDIA_CHANNEL,
    c.REACH: c.RF_CHANNEL,
    c.FREQUENCY: c.RF_CHANNEL,
    c.ORGANIC_MEDIA: c.ORGANIC_MEDIA_CHANNEL,
    c.ORGANIC_REACH: c.ORGANIC_RF_CHANNEL,
    c.ORGANIC_FREQUENCY: c.ORGANIC_RF_CHANNEL,
    c.NON_MEDIA_TREATMENTS: c.NON_MEDIA_CHANNEL,
}


@dataclasses.dataclass(frozen=True)
class _DeserializedTimeDimension:
  """Wrapper class for `TimeDimension` proto to provide utility methods during deserialization."""

  _time_dimension: marketing_pb.MarketingDataMetadata.TimeDimension

  def __post_init__(self):
    if not self._time_dimension.dates:
      raise ValueError("TimeDimension proto must have at least one date.")

  @functools.cached_property
  def date_coordinates(self) -> list[dt.date]:
    """Returns a list of date coordinates in this time dimension."""
    return [dt.date(d.year, d.month, d.day) for d in self._time_dimension.dates]

  @functools.cached_property
  def time_dimension_interval(self) -> date_interval_pb2.DateInterval:
    """Returns the `[start, end)` interval that spans this time dimension.

    This date interval spans all of the date coordinates in this time dimension.
    """
    date_intervals = time_record.convert_times_to_date_intervals(
        self.date_coordinates
    )
    return _get_date_interval_from_date_intervals(list(date_intervals.values()))


@dataclasses.dataclass(frozen=True)
class _DeserializedMetadata:
  """A container for parsed metadata from the `MarketingData` proto.

  Attributes:
    _metadata: The `MarketingDataMetadata` proto.
  """

  _metadata: marketing_pb.MarketingDataMetadata

  def __post_init__(self):
    # Evaluate the properties to trigger validation
    _ = self.time_dimension
    _ = self.media_time_dimension

  def _get_time_dimension(self, name: str) -> _DeserializedTimeDimension:
    """Helper method to get a specific TimeDimension proto by name."""
    for time_dimension in self._metadata.time_dimensions:
      if time_dimension.name == name:
        return _DeserializedTimeDimension(time_dimension)
    raise ValueError(f"No TimeDimension found with name '{name}' in metadata.")

  @functools.cached_property
  def time_dimension(self) -> _DeserializedTimeDimension:
    """Returns the TimeDimension with name 'time'."""
    return self._get_time_dimension(c.TIME)

  @functools.cached_property
  def media_time_dimension(self) -> _DeserializedTimeDimension:
    """Returns the TimeDimension with name 'media_time'."""
    return self._get_time_dimension(c.MEDIA_TIME)

  @functools.cached_property
  def channel_dimensions(self) -> Mapping[str, list[str]]:
    """Returns a mapping of channel dimension names to their corresponding channel coordinate names."""
    return {
        cd.name: list(cd.channels) for cd in self._metadata.channel_dimensions
    }

  @functools.cached_property
  def channel_types(self) -> Mapping[str, str | None]:
    """Returns a mapping of individual channel names to their types."""
    channel_coord_map = {}
    for name, channels in self.channel_dimensions.items():
      for channel in channels:
        channel_coord_map[channel] = _COORD_NAME_MAP.get(
            name,
        )
    return channel_coord_map


def _extract_data_array(
    serialized_data_points: Sequence[marketing_pb.MarketingDataPoint],
    data_extractor_fn,
    data_name,
) -> xr.DataArray | None:
  """Helper function to extract data into an `xr.DataArray`.

  Args:
    serialized_data_points: A Sequence of MarketingDataPoint protos.
    data_extractor_fn: A function that takes a data point and returns either a
      tuple of `(geo_id, time_str, value)`, or `None` if the data point should
      be skipped.
    data_name: The desired name for the `xr.DataArray`.

  Returns:
    An `xr.DataArray` containing the extracted data, or `None` if no data is
    found.
  """
  data_dict = {}  # (geo_id, time_str) -> value
  geo_ids = []
  times = []

  for data_point in serialized_data_points:
    extraction_result = data_extractor_fn(data_point)
    if extraction_result is None:
      continue

    geo_id, time_str, value = extraction_result

    # TODO: Enforce dimension uniqueness in Meridian.
    if geo_id not in geo_ids:
      geo_ids.append(geo_id)
    if time_str not in times:
      times.append(time_str)

    data_dict[(geo_id, time_str)] = value

  if not data_dict:
    return None

  data_values = np.array([
      [data_dict.get((geo_id, time), np.nan) for time in times]
      for geo_id in geo_ids
  ])

  return xr.DataArray(
      data=data_values,
      coords={
          c.GEO: geo_ids,
          c.TIME: times,
      },
      dims=(c.GEO, c.TIME),
      name=data_name,
  )


def _extract_3d_data_array(
    serialized_data_points: Sequence[marketing_pb.MarketingDataPoint],
    data_extractor_fn,
    data_name,
    third_dim_name,
    time_dim_name=c.TIME,
) -> xr.DataArray | None:
  """Helper function to extract data with 3 dimensions into an `xr.DataArray`.

  The first dimension is always `GEO`, and the second is the time dimension
  (default: `TIME`).

  Args:
      serialized_data_points: A sequence of MarketingDataPoint protos.
      data_extractor_fn: A function that takes a data point and returns either a
        tuple of `(geo_id, time_str, third_dim_key, value)`, or `None` if the
        data point should be skipped.
      data_name: The desired name for the `xr.DataArray`.
      third_dim_name: The name of the third dimension.
      time_dim_name: The name of the time dimension. Default is `TIME`.

  Returns:
      An `xr.DataArray` containing the extracted data, or `None` if no data is
      found.
  """
  data_dict = {}  # (geo_id, time_str, third_dim_key) -> value
  geo_ids = []
  times = []
  third_dim_keys = []

  for data_point in serialized_data_points:
    for extraction_result in data_extractor_fn(data_point):
      geo_id, time_str, third_dim_key, value = extraction_result

      if geo_id not in geo_ids:
        geo_ids.append(geo_id)
      if time_str not in times:
        times.append(time_str)
      if third_dim_key not in third_dim_keys:
        third_dim_keys.append(third_dim_key)

      # TODO: Enforce dimension uniqueness in Meridian.
      data_dict[(geo_id, time_str, third_dim_key)] = value

  if not data_dict:
    return None

  data_values = np.array([
      [
          [
              data_dict.get((geo_id, time, third_dim_key), np.nan)
              for third_dim_key in third_dim_keys
          ]
          for time in times
      ]
      for geo_id in geo_ids
  ])

  return xr.DataArray(
      data=data_values,
      coords={
          c.GEO: geo_ids,
          time_dim_name: times,
          third_dim_name: third_dim_keys,
      },
      dims=(c.GEO, time_dim_name, third_dim_name),
      name=data_name,
  )


def _get_date_interval_from_date_intervals(
    date_intervals: Sequence[date_interval_pb2.DateInterval],
) -> date_interval_pb2.DateInterval:
  """Gets the date interval based on the earliest start date and latest end date.

  Args:
      date_intervals: A list of DateInterval protos.

  Returns:
      A DateInterval representing the earliest start date and latest end date.
  """
  get_start_date = lambda interval: dt.date(
      interval.start_date.year,
      interval.start_date.month,
      interval.start_date.day,
  )
  get_end_date = lambda interval: dt.date(
      interval.end_date.year, interval.end_date.month, interval.end_date.day
  )

  min_start_date_interval = min(date_intervals, key=get_start_date)
  max_end_date_interval = max(date_intervals, key=get_end_date)

  return date_interval_pb2.DateInterval(
      start_date=date_pb2.Date(
          year=min_start_date_interval.start_date.year,
          month=min_start_date_interval.start_date.month,
          day=min_start_date_interval.start_date.day,
      ),
      end_date=date_pb2.Date(
          year=max_end_date_interval.end_date.year,
          month=max_end_date_interval.end_date.month,
          day=max_end_date_interval.end_date.day,
      ),
  )


class _InputDataSerializer:
  """Serializes an `InputData` container in Meridian model."""

  def __init__(self, input_data: meridian_input_data.InputData):
    self._input_data = input_data

  @property
  def _n_geos(self) -> int:
    return len(self._input_data.geo)

  @property
  def _n_times(self) -> int:
    return len(self._input_data.time)

  def __call__(self) -> marketing_pb.MarketingData:
    """Serializes the input data into a MarketingData proto."""
    marketing_proto = marketing_pb.MarketingData()
    # Use media_time since it covers larger range.
    times_to_date_intervals = time_record.convert_times_to_date_intervals(
        self._input_data.media_time.data
    )
    geos_and_times = itertools.product(
        self._input_data.geo.data, self._input_data.media_time.data
    )

    for geo, time in geos_and_times:
      data_point = self._serialize_data_point(
          geo,
          time,
          times_to_date_intervals,
      )
      marketing_proto.marketing_data_points.append(data_point)

    if self._input_data.media_spend is not None:
      if (
          not self._input_data.media_spend_has_geo_dimension
          and not self._input_data.media_spend_has_time_dimension
      ):
        marketing_proto.marketing_data_points.append(
            self._serialize_aggregated_media_spend_data_point(
                self._input_data.media_spend,
                times_to_date_intervals,
            )
        )
      elif (
          self._input_data.media_spend_has_geo_dimension
          != self._input_data.media_spend_has_time_dimension
      ):
        raise AssertionError(
            "Invalid input data: media_spend must either be fully granular"
            " (both geo and time dimensions) or fully aggregated (neither geo"
            " nor time dimensions)."
        )

    if self._input_data.rf_spend is not None:
      if (
          not self._input_data.rf_spend_has_geo_dimension
          and not self._input_data.rf_spend_has_time_dimension
      ):
        marketing_proto.marketing_data_points.append(
            self._serialize_aggregated_rf_spend_data_point(
                self._input_data.rf_spend, times_to_date_intervals
            )
        )
      elif (
          self._input_data.rf_spend_has_geo_dimension
          != self._input_data.rf_spend_has_time_dimension
      ):
        raise AssertionError(
            "Invalid input data: rf_spend must either be fully granular (both"
            " geo and time dimensions) or fully aggregated (neither geo nor"
            " time dimensions)."
        )

    marketing_proto.metadata.CopyFrom(self._serialize_metadata())

    return marketing_proto

  def _serialize_media_variables(
      self,
      geo: str,
      time: str,
      channel_dim_name: str,
      impressions_data_array: xr.DataArray,
      spend_data_array: xr.DataArray | None = None,
  ) -> list[marketing_pb.MediaVariable]:
    """Serializes media variables for a given geo and time.

    Args:
      geo: The geo ID.
      time: The time string.
      channel_dim_name: The name of the channel dimension.
      impressions_data_array: The DataArray containing impressions data.
        Expected dimensions: `(n_geos, n_media_times, n_channels)`.
      spend_data_array: The optional DataArray containing spend data. Expected
        dimensions are `(n_geos, n_times, n_media_channels)`.

    Returns:
      A list of MediaVariable protos.
    """
    media_variables = []
    for media_data in impressions_data_array.sel(geo=geo, media_time=time):
      channel = media_data[channel_dim_name].item()
      media_variable = marketing_pb.MediaVariable(
          channel_name=channel,
          scalar_metric=marketing_pb.ScalarMetric(
              name=c.IMPRESSIONS, value=media_data.item()
          ),
      )
      if spend_data_array is not None and time in spend_data_array.time:
        media_variable.media_spend = spend_data_array.sel(
            geo=geo, time=time, **{channel_dim_name: channel}
        ).item()
      media_variables.append(media_variable)
    return media_variables

  def _serialize_reach_frequency_variables(
      self,
      geo: str,
      time: str,
      channel_dim_name: str,
      reach_data_array: xr.DataArray,
      frequency_data_array: xr.DataArray,
      spend_data_array: xr.DataArray | None = None,
  ) -> list[marketing_pb.ReachFrequencyVariable]:
    """Serializes reach and frequency variables for a given geo and time.

    Iterates through the R&F channels separately, creating a MediaVariable
    for each. It's safe to assume that Meridian media channel names are
    unique across `media_data` and `reach_data`. This assumption is
    checked when an `InputData` is created in model training.

    Dimensions of `reach_data_array` and `frequency_data_array` are expected
    to be `(n_geos, n_media_times, n_rf_channels)`.

    Args:
      geo: The geo ID.
      time: The time string.
      channel_dim_name: The name of the channel dimension (e.g., 'rf_channel').
      reach_data_array: The DataArray containing reach data.
      frequency_data_array: The DataArray containing frequency data.
      spend_data_array: The optional DataArray containing spend data.

    Returns:
      A list of MediaVariable protos.
    """
    rf_variables = []
    for reach_data in reach_data_array.sel(geo=geo, media_time=time):
      reach_value = reach_data.item()
      channel = reach_data[channel_dim_name].item()
      frequency_value = frequency_data_array.sel(
          geo=geo,
          media_time=time,
          **{channel_dim_name: channel},
      ).item()
      rf_variable = marketing_pb.ReachFrequencyVariable(
          channel_name=channel,
          reach=int(reach_value),
          average_frequency=frequency_value,
      )
      if spend_data_array is not None and time in spend_data_array.time:
        rf_variable.spend = spend_data_array.sel(
            geo=geo, time=time, **{channel_dim_name: channel}
        ).item()
      rf_variables.append(rf_variable)
    return rf_variables

  def _serialize_non_media_treatment_variables(
      self, geo: str, time: str
  ) -> list[marketing_pb.NonMediaTreatmentVariable]:
    """Serializes non-media treatment variables for a given geo and time.

    Args:
      geo: The geo ID.
      time: The time string.

    Returns:
      A list of NonMediaTreatmentVariable protos.
    """
    non_media_treatment_variables = []
    if (
        self._input_data.non_media_treatments is not None
        and geo in self._input_data.non_media_treatments.geo
        and time in self._input_data.non_media_treatments.time
    ):
      for non_media_treatment_data in self._input_data.non_media_treatments.sel(
          geo=geo, time=time
      ):
        non_media_treatment_variables.append(
            marketing_pb.NonMediaTreatmentVariable(
                name=non_media_treatment_data[c.NON_MEDIA_CHANNEL].item(),
                value=non_media_treatment_data.item(),
            )
        )
    return non_media_treatment_variables

  def _serialize_data_point(
      self,
      geo: str,
      time: str,
      times_to_date_intervals: Mapping[str, date_interval_pb2.DateInterval],
  ) -> marketing_pb.MarketingDataPoint:
    """Serializes a MarketingDataPoint proto for a given geo and time."""
    data_point = marketing_pb.MarketingDataPoint(
        geo_info=marketing_pb.GeoInfo(
            geo_id=geo,
            population=round(self._input_data.population.sel(geo=geo).item()),
        ),
        date_interval=times_to_date_intervals.get(time),
    )

    if self._input_data.controls is not None:
      if time in self._input_data.controls.time:
        for control_data in self._input_data.controls.sel(geo=geo, time=time):
          data_point.control_variables.add(
              name=control_data.control_variable.item(),
              value=control_data.item(),
          )

    if self._input_data.media is not None:
      if (
          self._input_data.media_spend_has_geo_dimension
          and self._input_data.media_spend_has_time_dimension
      ):
        spend_data_array = self._input_data.media_spend
      else:
        # Aggregated spend data is serialized in a separate data point.
        spend_data_array = None
      media_variables = self._serialize_media_variables(
          geo,
          time,
          c.MEDIA_CHANNEL,
          self._input_data.media,
          spend_data_array,
      )
      data_point.media_variables.extend(media_variables)

    if (
        self._input_data.reach is not None
        and self._input_data.frequency is not None
    ):
      if (
          self._input_data.rf_spend_has_geo_dimension
          and self._input_data.rf_spend_has_time_dimension
      ):
        rf_spend_data_array = self._input_data.rf_spend
      else:
        # Aggregated spend data is serialized in a separate data point.
        rf_spend_data_array = None
      rf_variables = self._serialize_reach_frequency_variables(
          geo,
          time,
          c.RF_CHANNEL,
          self._input_data.reach,
          self._input_data.frequency,
          rf_spend_data_array,
      )
      data_point.reach_frequency_variables.extend(rf_variables)

    if self._input_data.organic_media is not None:
      organic_media_variables = self._serialize_media_variables(
          geo, time, c.ORGANIC_MEDIA_CHANNEL, self._input_data.organic_media
      )
      data_point.media_variables.extend(organic_media_variables)

    if (
        self._input_data.organic_reach is not None
        and self._input_data.organic_frequency is not None
    ):
      organic_rf_variables = self._serialize_reach_frequency_variables(
          geo,
          time,
          c.ORGANIC_RF_CHANNEL,
          self._input_data.organic_reach,
          self._input_data.organic_frequency,
      )
      data_point.reach_frequency_variables.extend(organic_rf_variables)

    non_media_treatment_variables = (
        self._serialize_non_media_treatment_variables(geo, time)
    )
    data_point.non_media_treatment_variables.extend(
        non_media_treatment_variables
    )

    if time in self._input_data.kpi.time:
      kpi_proto = self._make_kpi_proto(geo, time)
      data_point.kpi.CopyFrom(kpi_proto)

    return data_point

  def _serialize_aggregated_media_spend_data_point(
      self,
      spend_data_array: xr.DataArray,
      times_to_date_intervals: Mapping[str, date_interval_pb2.DateInterval],
  ) -> marketing_pb.MarketingDataPoint:
    """Serializes and appends a data point for aggregated spend."""
    spend_data_point = marketing_pb.MarketingDataPoint()
    date_interval = _get_date_interval_from_date_intervals(
        list(times_to_date_intervals.values())
    )
    spend_data_point.date_interval.CopyFrom(date_interval)

    for channel_name in spend_data_array.coords[c.MEDIA_CHANNEL].values:
      spend_value = spend_data_array.sel(
          **{c.MEDIA_CHANNEL: channel_name}
      ).item()
      spend_data_point.media_variables.add(
          channel_name=channel_name, media_spend=spend_value
      )

    return spend_data_point

  def _serialize_aggregated_rf_spend_data_point(
      self,
      spend_data_array: xr.DataArray,
      times_to_date_intervals: Mapping[str, date_interval_pb2.DateInterval],
  ) -> marketing_pb.MarketingDataPoint:
    """Serializes and appends a data point for aggregated spend."""
    spend_data_point = marketing_pb.MarketingDataPoint()
    date_interval = _get_date_interval_from_date_intervals(
        list(times_to_date_intervals.values())
    )
    spend_data_point.date_interval.CopyFrom(date_interval)

    for channel_name in spend_data_array.coords[c.RF_CHANNEL].values:
      spend_value = spend_data_array.sel(**{c.RF_CHANNEL: channel_name}).item()
      spend_data_point.reach_frequency_variables.add(
          channel_name=channel_name, spend=spend_value
      )

    return spend_data_point

  def _serialize_time_dimensions(
      self, name: str, time_data: xr.DataArray
  ) -> marketing_pb.MarketingDataMetadata.TimeDimension:
    """Creates a TimeDimension message."""
    time_dimensions = marketing_pb.MarketingDataMetadata.TimeDimension(
        name=name
    )
    for date in time_data.values:
      date_obj = dt.datetime.strptime(date, c.DATE_FORMAT).date()
      time_dimensions.dates.add(
          year=date_obj.year, month=date_obj.month, day=date_obj.day
      )
    return time_dimensions

  def _serialize_channel_dimensions(
      self, channel_data: xr.DataArray | None
  ) -> marketing_pb.MarketingDataMetadata.ChannelDimension | None:
    """Creates a ChannelDimension message if the corresponding attribute exists."""
    if channel_data is None:
      return None

    coord_name = _COORD_NAME_MAP.get(channel_data.name)
    if coord_name:
      return marketing_pb.MarketingDataMetadata.ChannelDimension(
          name=channel_data.name,
          channels=channel_data.coords[coord_name].values.tolist(),
      )
    else:
      # Make sure that all channel dimensions are handled.
      raise ValueError(f"Unknown channel data name: {channel_data.name}. ")

  def _serialize_metadata(self) -> marketing_pb.MarketingDataMetadata:
    """Serializes metadata from InputData to MarketingDataMetadata."""
    metadata = marketing_pb.MarketingDataMetadata()

    metadata.time_dimensions.append(
        self._serialize_time_dimensions(c.TIME, self._input_data.time)
    )
    metadata.time_dimensions.append(
        self._serialize_time_dimensions(
            c.MEDIA_TIME, self._input_data.media_time
        )
    )

    channel_data_arrays = [
        self._input_data.media,
        self._input_data.reach,
        self._input_data.frequency,
        self._input_data.organic_media,
        self._input_data.organic_reach,
        self._input_data.organic_frequency,
    ]

    for channel_data_array in channel_data_arrays:
      channel_names_message = self._serialize_channel_dimensions(
          channel_data_array
      )
      if channel_names_message:
        metadata.channel_dimensions.append(channel_names_message)

    if self._input_data.controls is not None:
      metadata.control_names.extend(
          self._input_data.controls.control_variable.values
      )

    if self._input_data.non_media_treatments is not None:
      metadata.non_media_treatment_names.extend(
          self._input_data.non_media_treatments.non_media_channel.values
      )

    metadata.kpi_type = self._input_data.kpi_type

    return metadata

  def _make_kpi_proto(self, geo: str, time: str) -> marketing_pb.Kpi:
    """Constructs a Kpi proto from the TrainedModel."""
    kpi_proto = marketing_pb.Kpi(name=self._input_data.kpi_type)
    # `kpi` and `revenue_per_kpi` dimensions: `(n_geos, n_times)`.
    if self._input_data.kpi_type == c.REVENUE:
      kpi_proto.revenue.CopyFrom(
          marketing_pb.Kpi.Revenue(
              value=self._input_data.kpi.sel(geo=geo, time=time).item()
          )
      )
    else:
      kpi_proto.non_revenue.CopyFrom(
          marketing_pb.Kpi.NonRevenue(
              value=self._input_data.kpi.sel(geo=geo, time=time).item()
          )
      )
      if self._input_data.revenue_per_kpi is not None:
        kpi_proto.non_revenue.revenue_per_kpi = (
            self._input_data.revenue_per_kpi.sel(geo=geo, time=time).item()
        )
    return kpi_proto


class _InputDataDeserializer:
  """Deserializes a `MarketingData` proto into a Meridian `InputData`."""

  def __init__(self, serialized: marketing_pb.MarketingData):
    self._serialized = serialized

  def __post_init__(self):
    if not self._serialized.HasField(sc.METADATA):
      raise ValueError(
          f"MarketingData proto is missing the '{sc.METADATA}' field."
      )

  @functools.cached_property
  def _metadata(self) -> _DeserializedMetadata:
    """Parses metadata and extracts time dimensions, channel dimensions, and channel type map."""
    return _DeserializedMetadata(self._serialized.metadata)

  def _extract_population(self) -> xr.DataArray:
    """Extracts population data from the serialized proto."""
    geo_populations = {}

    for data_point in self._serialized.marketing_data_points:
      geo_id = data_point.geo_info.geo_id
      if not geo_id:
        continue

      geo_populations[geo_id] = data_point.geo_info.population

    return xr.DataArray(
        coords={c.GEO: list(geo_populations.keys())},
        data=np.array(list(geo_populations.values())),
        name=c.POPULATION,
    )

  def _extract_kpi_type(self) -> str:
    """Extracts the kpi_type from the serialized proto."""
    kpi_type = None
    for data_point in self._serialized.marketing_data_points:
      if data_point.HasField(c.KPI):
        current_kpi_type = data_point.kpi.WhichOneof(c.TYPE)

        if kpi_type is None:
          kpi_type = current_kpi_type
        elif kpi_type != current_kpi_type:
          raise ValueError(
              "Inconsistent kpi_type found in the data. "
              f"Expected {kpi_type}, found {current_kpi_type}"
          )

    if kpi_type is None:
      raise ValueError("kpi_type not found in the data.")
    return kpi_type

  def _extract_geo_and_time(self, data_point) -> tuple[str | None, str]:
    """Extracts geo_id and time_str from a data_point."""
    geo_id = data_point.geo_info.geo_id
    start_date = data_point.date_interval.start_date
    time_str = dt.datetime(
        start_date.year, start_date.month, start_date.day
    ).strftime(c.DATE_FORMAT)
    return geo_id, time_str

  def _extract_kpi(self, kpi_type: str) -> xr.DataArray:
    """Extracts KPI data from the serialized proto."""

    def _kpi_extractor(data_point):
      if not data_point.HasField(c.KPI):
        return None

      geo_id, time_str = self._extract_geo_and_time(data_point)

      if data_point.kpi.WhichOneof(c.TYPE) != kpi_type:
        raise ValueError(
            "Inconsistent kpi_type found in the data. "
            f"Expected {kpi_type}, found"
            f" {data_point.kpi.WhichOneof(c.TYPE)}"
        )

      kpi_value = (
          data_point.kpi.revenue.value
          if kpi_type == c.REVENUE
          else data_point.kpi.non_revenue.value
      )
      return geo_id, time_str, kpi_value

    kpi = _extract_data_array(
        serialized_data_points=self._serialized.marketing_data_points,
        data_extractor_fn=_kpi_extractor,
        data_name=c.KPI,
    )

    if kpi is None:
      raise ValueError(f"{c.KPI} is not found in the data.")

    return kpi

  def _extract_revenue_per_kpi(self, kpi_type: str) -> xr.DataArray | None:
    """Extracts revenue per KPI data from the serialized proto."""

    if kpi_type == c.REVENUE:
      raise ValueError(
          f"{c.REVENUE_PER_KPI} is not applicable when kpi_type is {c.REVENUE}."
      )

    def _revenue_per_kpi_extractor(data_point):
      if not data_point.HasField(c.KPI):
        return None

      if not data_point.kpi.non_revenue.HasField(c.REVENUE_PER_KPI):
        return None

      geo_id, time_str = self._extract_geo_and_time(data_point)

      if data_point.kpi.WhichOneof(c.TYPE) != kpi_type:
        raise ValueError(
            "Inconsistent kpi_type found in the data. "
            f"Expected {kpi_type}, found"
            f" {data_point.kpi.WhichOneof(c.TYPE)}"
        )

      return geo_id, time_str, data_point.kpi.non_revenue.revenue_per_kpi

    return _extract_data_array(
        serialized_data_points=self._serialized.marketing_data_points,
        data_extractor_fn=_revenue_per_kpi_extractor,
        data_name=c.REVENUE_PER_KPI,
    )

  def _extract_controls(self) -> xr.DataArray | None:
    """Extracts control variables data from the serialized proto, if any."""

    def _controls_extractor(data_point):
      if not data_point.control_variables:
        return None

      geo_id, time_str = self._extract_geo_and_time(data_point)

      for control_variable in data_point.control_variables:
        control_name = control_variable.name
        control_value = control_variable.value
        yield geo_id, time_str, control_name, control_value

    return _extract_3d_data_array(
        serialized_data_points=self._serialized.marketing_data_points,
        data_extractor_fn=_controls_extractor,
        data_name=c.CONTROLS,
        third_dim_name=c.CONTROL_VARIABLE,
    )

  def _extract_media(self) -> xr.DataArray | None:
    """Extracts media variables data from the serialized proto."""

    def _media_extractor(data_point):
      geo_id, time_str = self._extract_geo_and_time(data_point)

      if not geo_id:
        return None

      for media_variable in data_point.media_variables:
        channel_name = media_variable.channel_name
        if self._metadata.channel_types.get(channel_name) != c.MEDIA_CHANNEL:
          continue

        media_value = media_variable.scalar_metric.value
        yield geo_id, time_str, channel_name, media_value

    return _extract_3d_data_array(
        serialized_data_points=self._serialized.marketing_data_points,
        data_extractor_fn=_media_extractor,
        data_name=c.MEDIA,
        third_dim_name=c.MEDIA_CHANNEL,
        time_dim_name=c.MEDIA_TIME,
    )

  def _extract_reach(self) -> xr.DataArray | None:
    """Extracts reach data from the serialized proto."""

    def _reach_extractor(data_point):
      geo_id, time_str = self._extract_geo_and_time(data_point)

      if not geo_id:
        return None

      for rf_variable in data_point.reach_frequency_variables:
        channel_name = rf_variable.channel_name
        if self._metadata.channel_types.get(channel_name) != c.RF_CHANNEL:
          continue

        reach_value = rf_variable.reach
        yield geo_id, time_str, channel_name, reach_value

    return _extract_3d_data_array(
        serialized_data_points=self._serialized.marketing_data_points,
        data_extractor_fn=_reach_extractor,
        data_name=c.REACH,
        third_dim_name=c.RF_CHANNEL,
        time_dim_name=c.MEDIA_TIME,
    )

  def _extract_frequency(self) -> xr.DataArray | None:
    """Extracts frequency data from the serialized proto."""

    def _frequency_extractor(data_point):
      geo_id, time_str = self._extract_geo_and_time(data_point)

      if not geo_id:
        return None

      for rf_variable in data_point.reach_frequency_variables:
        channel_name = rf_variable.channel_name
        if self._metadata.channel_types.get(channel_name) != c.RF_CHANNEL:
          continue

        frequency_value = rf_variable.average_frequency
        yield geo_id, time_str, channel_name, frequency_value

    return _extract_3d_data_array(
        serialized_data_points=self._serialized.marketing_data_points,
        data_extractor_fn=_frequency_extractor,
        data_name=c.FREQUENCY,
        third_dim_name=c.RF_CHANNEL,
        time_dim_name=c.MEDIA_TIME,
    )

  def _extract_organic_media(self) -> xr.DataArray | None:
    """Extracts organic media variables data from the serialized proto."""

    def _organic_media_extractor(data_point):
      geo_id, time_str = self._extract_geo_and_time(data_point)

      if not geo_id:
        return None

      for media_variable in data_point.media_variables:
        channel_name = media_variable.channel_name
        if (
            self._metadata.channel_types.get(channel_name)
            != c.ORGANIC_MEDIA_CHANNEL
        ):
          continue

        media_value = media_variable.scalar_metric.value
        yield geo_id, time_str, channel_name, media_value

    return _extract_3d_data_array(
        serialized_data_points=self._serialized.marketing_data_points,
        data_extractor_fn=_organic_media_extractor,
        data_name=c.ORGANIC_MEDIA,
        third_dim_name=c.ORGANIC_MEDIA_CHANNEL,
        time_dim_name=c.MEDIA_TIME,
    )

  def _extract_organic_reach(self) -> xr.DataArray | None:
    """Extracts organic reach data from the serialized proto."""

    def _organic_reach_extractor(data_point):
      geo_id, time_str = self._extract_geo_and_time(data_point)

      if not geo_id:
        return None

      for rf_variable in data_point.reach_frequency_variables:
        channel_name = rf_variable.channel_name
        if (
            self._metadata.channel_types.get(channel_name)
            != c.ORGANIC_RF_CHANNEL
        ):
          continue

        reach_value = rf_variable.reach
        yield geo_id, time_str, channel_name, reach_value

    return _extract_3d_data_array(
        serialized_data_points=self._serialized.marketing_data_points,
        data_extractor_fn=_organic_reach_extractor,
        data_name=c.ORGANIC_REACH,
        third_dim_name=c.ORGANIC_RF_CHANNEL,
        time_dim_name=c.MEDIA_TIME,
    )

  def _extract_organic_frequency(self) -> xr.DataArray | None:
    """Extracts organic frequency data from the serialized proto."""

    def _organic_frequency_extractor(data_point):
      geo_id, time_str = self._extract_geo_and_time(data_point)

      if not geo_id:
        return None

      for rf_variable in data_point.reach_frequency_variables:
        channel_name = rf_variable.channel_name
        if (
            self._metadata.channel_types.get(channel_name)
            != c.ORGANIC_RF_CHANNEL
        ):
          continue

        frequency_value = rf_variable.average_frequency
        yield geo_id, time_str, channel_name, frequency_value

    return _extract_3d_data_array(
        serialized_data_points=self._serialized.marketing_data_points,
        data_extractor_fn=_organic_frequency_extractor,
        data_name=c.ORGANIC_FREQUENCY,
        third_dim_name=c.ORGANIC_RF_CHANNEL,
        time_dim_name=c.MEDIA_TIME,
    )

  def _extract_granular_media_spend(
      self,
      data_points_with_spend: list[marketing_pb.MarketingDataPoint],
  ) -> xr.DataArray | None:
    """Extracts granular spend data.

    Args:
      data_points_with_spend: List of MarketingDataPoint protos with spend data.

    Returns:
      An xr.DataArray with granular spend data or None if no data found.
    """

    def _granular_spend_extractor(data_point):
      geo_id, time_str = self._extract_geo_and_time(data_point)
      for media_variable in data_point.media_variables:
        if (
            media_variable.HasField(c.MEDIA_SPEND)
            and self._metadata.channel_types.get(media_variable.channel_name)
            == c.MEDIA_CHANNEL
        ):
          yield geo_id, time_str, media_variable.channel_name, media_variable.media_spend

    return _extract_3d_data_array(
        serialized_data_points=data_points_with_spend,
        data_extractor_fn=_granular_spend_extractor,
        data_name=c.MEDIA_SPEND,
        third_dim_name=c.MEDIA_CHANNEL,
        time_dim_name=c.TIME,
    )

  def _extract_granular_rf_spend(
      self,
      data_points_with_spend: list[marketing_pb.MarketingDataPoint],
  ) -> xr.DataArray | None:
    """Extracts granular spend data.

    Args:
      data_points_with_spend: List of MarketingDataPoint protos with spend data.

    Returns:
      An xr.DataArray with granular spend data or None if no data found.
    """

    def _granular_spend_extractor(data_point):
      geo_id, time_str = self._extract_geo_and_time(data_point)
      for rf_variable in data_point.reach_frequency_variables:
        if (
            rf_variable.HasField(c.SPEND)
            and self._metadata.channel_types.get(rf_variable.channel_name)
            == c.RF_CHANNEL
        ):
          yield geo_id, time_str, rf_variable.channel_name, rf_variable.spend

    return _extract_3d_data_array(
        serialized_data_points=data_points_with_spend,
        data_extractor_fn=_granular_spend_extractor,
        data_name=c.RF_SPEND,
        third_dim_name=c.RF_CHANNEL,
        time_dim_name=c.TIME,
    )

  def _extract_aggregated_media_spend(
      self,
      data_points_with_spend: list[marketing_pb.MarketingDataPoint],
  ) -> xr.DataArray | None:
    """Extracts aggregated spend data.

    Args:
      data_points_with_spend: List of MarketingDataPoint protos with spend data.

    Returns:
      An xr.DataArray with aggregated spend data or None if no data found.
    """
    channel_names = self._metadata.channel_dimensions.get(c.MEDIA, [])
    channel_spend_map = {}

    for spend_data_point in data_points_with_spend:
      for media_variable in spend_data_point.media_variables:
        if (
            media_variable.channel_name in channel_names
            and media_variable.HasField(c.MEDIA_SPEND)
        ):
          channel_spend_map[media_variable.channel_name] = (
              media_variable.media_spend
          )

    if not channel_spend_map:
      return None

    return xr.DataArray(
        data=list(channel_spend_map.values()),
        coords={c.MEDIA_CHANNEL: list(channel_spend_map.keys())},
        dims=[c.MEDIA_CHANNEL],
        name=c.MEDIA_SPEND,
    )

  def _extract_aggregated_rf_spend(
      self,
      data_points_with_spend: list[marketing_pb.MarketingDataPoint],
  ) -> xr.DataArray | None:
    """Extracts aggregated spend data.

    Args:
      data_points_with_spend: List of MarketingDataPoint protos with spend data.

    Returns:
      An xr.DataArray with aggregated spend data or None if no data found.
    """
    channel_names = self._metadata.channel_dimensions.get(c.REACH, [])
    channel_spend_map = {}

    for spend_data_point in data_points_with_spend:
      for rf_variable in spend_data_point.reach_frequency_variables:
        if rf_variable.channel_name in channel_names and rf_variable.HasField(
            c.SPEND
        ):
          channel_spend_map[rf_variable.channel_name] = rf_variable.spend

    if not channel_spend_map:
      return None

    return xr.DataArray(
        data=list(channel_spend_map.values()),
        coords={c.RF_CHANNEL: list(channel_spend_map.keys())},
        dims=[c.RF_CHANNEL],
        name=c.RF_SPEND,
    )

  def _is_aggregated_spend_data_point(
      self, dp: marketing_pb.MarketingDataPoint
  ) -> bool:
    """Checks if a MarketingDataPoint with spend represents aggregated spend data.

    Args:
      dp: A marketing_pb.MarketingDataPoint representing a spend data point.

    Returns:
      True if the data point represents aggregated spend, False otherwise.
    """
    if not dp.HasField(sc.GEO_INFO) and self._metadata.media_time_dimension:
      media_time_interval = (
          self._metadata.media_time_dimension.time_dimension_interval
      )
      return (
          media_time_interval.start_date == dp.date_interval.start_date
          and media_time_interval.end_date == dp.date_interval.end_date
      )
    return False

  def _extract_media_spend(self) -> xr.DataArray | None:
    """Extracts media spend data from the serialized proto.

    Returns:
      An xr.DataArray with spend data or None if no data found.
    """
    # Filter data points relevant to spend based on channel type map
    media_channels = {
        channel
        for channel, metadata_channel_type in self._metadata.channel_types.items()
        if metadata_channel_type == c.MEDIA_CHANNEL
    }
    spend_data_points = [
        dp
        for dp in self._serialized.marketing_data_points
        if any(
            mv.HasField(c.MEDIA_SPEND) and mv.channel_name in media_channels
            for mv in dp.media_variables
        )
    ]

    if not spend_data_points:
      return None

    aggregated_spend_data_points = [
        dp
        for dp in spend_data_points
        if self._is_aggregated_spend_data_point(dp)
    ]

    if aggregated_spend_data_points:
      return self._extract_aggregated_media_spend(aggregated_spend_data_points)

    return self._extract_granular_media_spend(spend_data_points)

  def _extract_rf_spend(self) -> xr.DataArray | None:
    """Extracts reach and frequency spend data from the serialized proto.

    Returns:
      An xr.DataArray with spend data or None if no data found.
    """
    # Filter data points relevant to spend based on channel type map
    rf_channels = {
        channel
        for channel, metadata_channel_type in self._metadata.channel_types.items()
        if metadata_channel_type == c.RF_CHANNEL
    }
    spend_data_points = [
        dp
        for dp in self._serialized.marketing_data_points
        if any(
            mv.HasField(c.SPEND) and mv.channel_name in rf_channels
            for mv in dp.reach_frequency_variables
        )
    ]

    if not spend_data_points:
      return None

    aggregated_spend_data_points = [
        dp
        for dp in spend_data_points
        if self._is_aggregated_spend_data_point(dp)
    ]

    if aggregated_spend_data_points:
      return self._extract_aggregated_rf_spend(aggregated_spend_data_points)

    return self._extract_granular_rf_spend(spend_data_points)

  def _extract_non_media_treatments(self) -> xr.DataArray | None:
    """Extracts non-media treatment variables data from the serialized proto."""

    def _non_media_treatments_extractor(data_point):
      if not data_point.non_media_treatment_variables:
        return None

      geo_id, time_str = self._extract_geo_and_time(data_point)

      for (
          non_media_treatment_variable
      ) in data_point.non_media_treatment_variables:
        treatment_name = non_media_treatment_variable.name
        treatment_value = non_media_treatment_variable.value
        yield geo_id, time_str, treatment_name, treatment_value

    non_media_treatments_data_array = _extract_3d_data_array(
        serialized_data_points=self._serialized.marketing_data_points,
        data_extractor_fn=_non_media_treatments_extractor,
        data_name=c.NON_MEDIA_TREATMENTS,
        third_dim_name=c.NON_MEDIA_CHANNEL,
    )

    return non_media_treatments_data_array

  def __call__(self) -> meridian_input_data.InputData:
    """Converts the `MarketingData` proto to a Meridian `InputData`."""
    kpi_type = self._extract_kpi_type()
    return meridian_input_data.InputData(
        kpi=self._extract_kpi(kpi_type),
        kpi_type=kpi_type,
        controls=self._extract_controls(),
        population=self._extract_population(),
        revenue_per_kpi=(
            self._extract_revenue_per_kpi(kpi_type)
            if kpi_type == c.NON_REVENUE
            else None
        ),
        media=self._extract_media(),
        media_spend=self._extract_media_spend(),
        reach=self._extract_reach(),
        frequency=self._extract_frequency(),
        rf_spend=self._extract_rf_spend(),
        organic_media=self._extract_organic_media(),
        organic_reach=self._extract_organic_reach(),
        organic_frequency=self._extract_organic_frequency(),
        non_media_treatments=self._extract_non_media_treatments(),
    )


class MarketingDataSerde(
    serde.Serde[marketing_pb.MarketingData, meridian_input_data.InputData]
):
  """Serializes and deserializes an `InputData` container in Meridian."""

  def serialize(
      self, obj: meridian_input_data.InputData
  ) -> marketing_pb.MarketingData:
    """Serializes the given Meridian input data into a `MarketingData` proto."""
    return _InputDataSerializer(obj)()

  def deserialize(
      self, serialized: marketing_pb.MarketingData, serialized_version: str = ""
  ) -> meridian_input_data.InputData:
    """Deserializes the given `MarketingData` proto.

    Args:
      serialized: The serialized `MarketingData` proto.
      serialized_version: The version of the serialized model. This is used to
        handle changes in deserialization logic across different versions.

    Returns:
      A Meridian input data container.
    """
    return _InputDataDeserializer(serialized)()



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains an object to store prior distributions.

The `PriorDistribution` object contains distributions for various parameters
used by the Meridian model object.
"""

from __future__ import annotations

from collections.abc import MutableMapping, Sequence
import dataclasses
from typing import Any
import warnings

from meridian import backend
from meridian import constants
import numpy as np


__all__ = [
    'IndependentMultivariateDistribution',
    'PriorDistribution',
    'distributions_are_equal',
    'lognormal_dist_from_mean_std',
    'lognormal_dist_from_range',
]


@dataclasses.dataclass(kw_only=True)
class PriorDistribution:
  """Contains prior distributions for each model parameter.

  PriorDistribution is a utility class for Meridian. The required shapes of the
  arguments to `PriorDistribution` depend on the modeling options and data
  shapes passed to Meridian. For example, `ec_m` is a parameter that represents
  the half-saturation for each media channel. The `ec_m` argument must have
  either `batch_shape=[]` or `batch_shape` equal to the number of media
  channels. In the case of the former, each media channel gets the same prior.

  An error is raised upon Meridian construction if any prior distribution
  has a shape that cannot be broadcast to the shape designated by the model
  specification.

  The parameter batch shapes are as follows:

  | Parameter             | Batch shape                |
  |-----------------------|----------------------------|
  | `knot_values`         | `n_knots`                  |
  | `tau_g_excl_baseline` | `n_geos - 1`               |
  | `beta_m`              | `n_media_channels`         |
  | `beta_rf`             | `n_rf_channels`            |
  | `beta_om`             | `n_organic_media_channels` |
  | `beta_orf`            | `n_organic_rf_channels`    |
  | `eta_m`               | `n_media_channels`         |
  | `eta_rf`              | `n_rf_channels`            |
  | `eta_om`              | `n_organic_media_channels` |
  | `eta_orf`             | `n_organic_rf_channels`    |
  | `gamma_c`             | `n_controls`               |
  | `gamma_n`             | `n_non_media_channels`     |
  | `xi_c`                | `n_controls`               |
  | `xi_n`                | `n_non_media_channels`     |
  | `alpha_m`             | `n_media_channels`         |
  | `alpha_rf`            | `n_rf_channels`            |
  | `alpha_om`            | `n_organic_media_channels` |
  | `alpha_orf`           | `n_organic_rf_channels`    |
  | `ec_m`                | `n_media_channels`         |
  | `ec_rf`               | `n_rf_channels`            |
  | `ec_om`               | `n_organic_media_channels` |
  | `ec_orf`              | `n_organic_rf_channels`    |
  | `slope_m`             | `n_media_channels`         |
  | `slope_rf`            | `n_rf_channels`            |
  | `slope_om`            | `n_organic_media_channels` |
  | `slope_orf`           | `n_organic_rf_channels`    |
  | `sigma`               | (σ)                        |
  | `roi_m`               | `n_media_channels`         |
  | `roi_rf`              | `n_rf_channels`            |
  | `mroi_m`              | `n_media_channels`         |
  | `mroi_rf`             | `n_rf_channels`            |
  | `contribution_m`      | `n_media_channels`         |
  | `contribution_rf`     | `n_rf_channels`            |
  | `contribution_om`     | `n_organic_media_channels` |
  | `contribution_orf`    | `n_organic_f_channels`     |
  | `contribution_n`      | `n_non_media_channels`     |

  (σ) `n_geos` if `unique_sigma_for_each_geo`, otherwise this is `1`

  Attributes:
    knot_values: Prior distribution on knots for time effects. Default
      distribution is `Normal(0.0, 5.0)`.
    tau_g_excl_baseline: Prior distribution on geo effects, which represent the
      average KPI of each geo relative to the baseline geo. This parameter is
      broadcast to a vector of length `n_geos - 1`, preserving the geo order and
      excluding the `baseline_geo`. After sampling, `Meridian.inference_data`
      includes a modified version of this parameter called `tau_g`, which has
      length `n_geos` and contains a zero in the position corresponding to
      `baseline_geo`. Meridian ignores this distribution if `n_geos = 1`.
      Default distribution is `Normal(0.0, 5.0)`.
    beta_m: Prior distribution on a parameter for the hierarchical distribution
      of geo-level media effects for impression media channels (`beta_gm`). When
      `media_effects_dist` is set to `'normal'`, it is the hierarchical mean.
      When `media_effects_dist` is set to `'log_normal'`, it is the hierarchical
      parameter for the mean of the underlying, log-transformed, `Normal`
      distribution. Meridian ignores this distribution if
      `paid_media_prior_type` is `'roi'` or `'mroi'`, and uses the `roi_m` or
      `mroi_m` prior instead. Default distribution is `HalfNormal(5.0)`.
    beta_rf: Prior distribution on a parameter for the hierarchical distribution
      of geo-level media effects for reach and frequency media channels
      (`beta_grf`). When `media_effects_dist` is set to `'normal'`, it is the
      hierarchical mean. When `media_effects_dist` is set to `'log_normal'`, it
      is the hierarchical parameter for the mean of the underlying,
      log-transformed, `Normal` distribution. Meridian ignores this distribution
      if `paid_media_prior_type` is `'roi'` or `'mroi'`, and uses the `roi_m` or
      `mroi_rf` prior instead. Default distribution is `HalfNormal(5.0)`.
    beta_om: Prior distribution on a parameter for the hierarchical distribution
      of geo-level media effects for organic media channels (`beta_gom`). When
      `media_effects_dist` is set to `'normal'`, it is the hierarchical mean.
      When `media_effects_dist` is set to `'log_normal'`, it is the hierarchical
      parameter for the mean of the underlying, log-transformed, `Normal`
      distribution. Default distribution is `HalfNormal(5.0)`.
    beta_orf: Prior distribution on a parameter for the hierarchical
      distribution of geo-level media effects for organic reach and frequency
      media channels (`beta_gorf`). When `media_effects_dist` is set to
      `'normal'`, it is the hierarchical mean. When `media_effects_dist` is set
      to `'log_normal'`, it is the hierarchical parameter for the mean of the
      underlying, log-transformed, `Normal` distribution. Default distribution
      is `HalfNormal(5.0)`.
    eta_m: Prior distribution on a parameter for the hierarchical distribution
      of geo-level media effects for impression media channels (`beta_gm`). When
      `media_effects_dist` is set to `'normal'`, it is the hierarchical standard
      deviation. When `media_effects_dist` is set to `'log_normal'` it is the
      hierarchical parameter for the standard deviation of the underlying,
      log-transformed, `Normal` distribution. Default distribution is
      `HalfNormal(1.0)`.
    eta_rf: Prior distribution on a parameter for the hierarchical distribution
      of geo-level media effects for RF media channels (`beta_grf`). When
      `media_effects_dist` is set to `'normal'`, it is the hierarchical standard
      deviation. When `media_effects_dist` is set to `'log_normal'` it is the
      hierarchical parameter for the standard deviation of the underlying,
      log-transformed, `Normal` distribution. Default distribution is
      `HalfNormal(1.0)`.
    eta_om: Prior distribution on a parameter for the hierarchical distribution
      of geo-level media effects for organic media channels (`beta_gom`). When
      `media_effects_dist` is set to `'normal'`, it is the hierarchical standard
      deviation. When `media_effects_dist` is set to `'log_normal'` it is the
      hierarchical parameter for the standard deviation of the underlying,
      log-transformed, `Normal` distribution. Default distribution is
      `HalfNormal(1.0)`.
    eta_orf: Prior distribution on a parameter for the hierarchical distribution
      of geo-level media effects for organic RF media channels (`beta_gorf`).
      When `media_effects_dist` is set to `'normal'`, it is the hierarchical
      standard deviation. When `media_effects_dist` is set to `'log_normal'` it
      is the hierarchical parameter for the standard deviation of the
      underlying, log-transformed, `Normal` distribution. Default distribution
      is `HalfNormal(1.0)`.
    gamma_c: Prior distribution on the hierarchical mean of `gamma_gc` which is
      the coefficient on control `c` for geo `g`. Hierarchy is defined over
      geos. Default distribution is `Normal(0.0, 5.0)`.
    gamma_n: Prior distribution on the hierarchical mean of `gamma_gn` which is
      the coefficient on non-media channel `n` for geo `g`. Hierarchy is defined
      over geos. Default distribution is `Normal(0.0, 5.0)`.
    xi_c: Prior distribution on the hierarchical standard deviation of
      `gamma_gc` which is the coefficient on control `c` for geo `g`. Hierarchy
      is defined over geos. Default distribution is `HalfNormal(5.0)`.
    xi_n: Prior distribution on the hierarchical standard deviation of
      `gamma_gn` which is the coefficient on non-media channel `n` for geo `g`.
      Hierarchy is defined over geos. Default distribution is `HalfNormal(5.0)`.
    alpha_m: Prior distribution on the Adstock decay parameter for media input.
      Default distribution is `Uniform(0.0, 1.0)`.
    alpha_rf: Prior distribution on the Adstock decay parameter for RF input.
      Default distribution is `Uniform(0.0, 1.0)`.
    alpha_om: Prior distribution on the Adstock decay parameter for organic
      media input. Default distribution is `Uniform(0.0, 1.0)`.
    alpha_orf: Prior distribution on the Adstock decay parameter for organic RF
      input. Default distribution is `Uniform(0.0, 1.0)`.
    ec_m: Prior distribution on the `half-saturation` Hill parameter for media
      input. Default distribution is `TruncatedNormal(0.8, 0.8, 0.1, 10)`.
    ec_rf: Prior distribution on the `half-saturation` Hill parameter for RF
      input. Default distribution is `TransformedDistribution(LogNormal(0.7,
      0.4), Shift(0.1))`.
    ec_om: Prior distribution on the `half-saturation` Hill parameter for
      organic media input. Default distribution is `TruncatedNormal(0.8, 0.8,
      0.1, 10)`.
    ec_orf: Prior distribution on the `half-saturation` Hill parameter for
      organic RF input. Default distribution is `TransformedDistribution(
      LogNormal(0.7, 0.4), Shift(0.1))`.
    slope_m: Prior distribution on the `slope` Hill parameter for media input.
      Default distribution is `Deterministic(1.0)`.
    slope_rf: Prior distribution on the `slope` Hill parameter for RF input.
      Default distribution is `LogNormal(0.7, 0.4)`.
    slope_om: Prior distribution on the `slope` Hill parameter for organic media
      input. Default distribution is `Deterministic(1.0)`.
    slope_orf: Prior distribution on the `slope` Hill parameter for organic RF
      input. Default distribution is `LogNormal(0.7, 0.4)`.
    sigma: Prior distribution on the standard deviation of noise. Default
      distribution is `HalfNormal(5.0)`.
    roi_m: Prior distribution on the ROI of each media channel. This parameter
      is only used when `paid_media_prior_type` is `'roi'`, in which case
      `beta_m` is calculated as a deterministic function of `roi_m`, `alpha_m`,
      `ec_m`, `slope_m`, and the spend associated with each media channel.
      Default distribution is `LogNormal(0.2, 0.9)`. When `kpi_type` is
      `'non_revenue'` and `revenue_per_kpi` is not provided, ROI is interpreted
      as incremental KPI units per monetary unit spent. In this case, the
      default value for `roi_m` and `roi_rf` will be ignored and a common ROI
      prior will be assigned to all channels to achieve a target mean and
      standard deviation on the total media contribution.
    roi_rf: Prior distribution on the ROI of each Reach & Frequency channel.
      This parameter is only used when `paid_media_prior_type` is `'roi'`, in
      which case `beta_rf` is calculated as a deterministic function of
      `roi_rf`, `alpha_rf`, `ec_rf`, `slope_rf`, and the spend associated with
      each RF channel. Default distribution is `LogNormal(0.2, 0.9)`. When
      `kpi_type` is `'non_revenue'` and `revenue_per_kpi` is not provided, ROI
      is interpreted as incremental KPI units per monetary unit spent. In this
      case, the default value for `roi_m` and `roi_rf` will be ignored and a
      common ROI prior will be assigned to all channels to achieve a target mean
      and standard deviation on the total media contribution.
    mroi_m: Prior distribution on the mROI of each media channel. This parameter
      is only used when `paid_media_prior_type` is `'mroi'`, in which case
      `beta_m` is calculated as a deterministic function of `mroi_m`, `alpha_m`,
      `ec_m`, `slope_m`, and the spend associated with each media channel.
      Default distribution is `LogNormal(0.0, 0.5)`. When `kpi_type` is
      `'non_revenue'` and `revenue_per_kpi` is not provided, mROI is interpreted
      as the marginal incremental KPI units per monetary unit spent. In this
      case, a default distribution is not provided, so the user must specify it.
    mroi_rf: Prior distribution on the mROI of each Reach & Frequency channel.
      This parameter is only used when `paid_media_prior_type` is `'mroi'`, in
      which case `beta_rf` is calculated as a deterministic function of
      `mroi_rf`, `alpha_rf`, `ec_rf`, `slope_rf`, and the spend associated with
      each media channel. Default distribution is `LogNormal(0.0, 0.5)`. When
      `kpi_type` is `'non_revenue'` and `revenue_per_kpi` is not provided, mROI
      is interpreted as the marginal incremental KPI units per monetary unit
      spent. In this case, a default distribution is not provided, so the user
      must specify it.
    contribution_m: Prior distribution on the contribution of each media channel
      as a percentage of total outcome. This parameter is only used when
      `paid_media_prior_type` is `'contribution'`, in which case `beta_m` is
      calculated as a deterministic function of `contribution_m`, `alpha_m`,
      `ec_m`, `slope_m`, and the total outcome. Default distribution is
      `Beta(1.0, 99.0)`.
    contribution_rf: Prior distribution on the contribution of each Reach &
      Frequency channel as a percentage of total outcome. This parameter is only
      used when `paid_media_prior_type` is `'contribution'`, in which case
      `beta_rf` is calculated as a deterministic function of `contribution_rf`,
      `alpha_rf`, `ec_rf`, `slope_rf`, and the total outcome. Default
      distribution is `Beta(1.0, 99.0)`.
    contribution_om: Prior distribution on the contribution of each organic
      media channel as a percentage of total outcome. This parameter is only
      used when `organic_media_prior_type` is `'contribution'`, in which case
      `beta_om` is calculated as a deterministic function of `contribution_om`,
      `alpha_om`, `ec_om`, `slope_om`, and the total outcome. Default
      distribution is `Beta(1.0, 99.0)`.
    contribution_orf: Prior distribution on the contribution of each organic
      Reach & Frequency channel as a percentage of total outcome. This parameter
      is only used when `organic_media_prior_type` is `'contribution'`, in which
      case `beta_orf` is calculated as a deterministic function of
      `contribution_orf`, `alpha_orf`, `ec_orf`, `slope_orf`, and the total
      outcome. Default distribution is `Beta(1.0, 99.0)`.
    contribution_n: Prior distribution on the contribution of each non-media
      treatment channel as a percentage of total outcome. This parameter is only
      used when `non_media_treatment_prior_type` is `'contribution'`, in which
      case `gamma_n` is calculated as a deterministic function of
      `contribution_n` and the total outcome. Default distribution is
      `TruncatedNormal(0.0, 0.1, -1.0, 1.0)`.
  """

  knot_values: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.Normal(
          0.0, 5.0, name=constants.KNOT_VALUES
      ),
  )
  tau_g_excl_baseline: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.Normal(
          0.0, 5.0, name=constants.TAU_G_EXCL_BASELINE
      ),
  )
  beta_m: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.HalfNormal(
          5.0, name=constants.BETA_M
      ),
  )
  beta_rf: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.HalfNormal(
          5.0, name=constants.BETA_RF
      ),
  )
  beta_om: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.HalfNormal(
          5.0, name=constants.BETA_OM
      ),
  )
  beta_orf: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.HalfNormal(
          5.0, name=constants.BETA_ORF
      ),
  )
  eta_m: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.HalfNormal(1.0, name=constants.ETA_M),
  )
  eta_rf: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.HalfNormal(
          1.0, name=constants.ETA_RF
      ),
  )
  eta_om: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.HalfNormal(
          1.0, name=constants.ETA_OM
      ),
  )
  eta_orf: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.HalfNormal(
          1.0, name=constants.ETA_ORF
      ),
  )
  gamma_c: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.Normal(
          0.0, 5.0, name=constants.GAMMA_C
      ),
  )
  gamma_n: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.Normal(
          0.0, 5.0, name=constants.GAMMA_N
      ),
  )
  xi_c: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.HalfNormal(5.0, name=constants.XI_C),
  )
  xi_n: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.HalfNormal(5.0, name=constants.XI_N),
  )
  alpha_m: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.Uniform(
          0.0, 1.0, name=constants.ALPHA_M
      ),
  )
  alpha_rf: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.Uniform(
          0.0, 1.0, name=constants.ALPHA_RF
      ),
  )
  alpha_om: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.Uniform(
          0.0, 1.0, name=constants.ALPHA_OM
      ),
  )
  alpha_orf: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.Uniform(
          0.0, 1.0, name=constants.ALPHA_ORF
      ),
  )
  ec_m: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.TruncatedNormal(
          0.8, 0.8, 0.1, 10, name=constants.EC_M
      ),
  )
  ec_rf: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.TransformedDistribution(
          backend.tfd.LogNormal(0.7, 0.4),
          backend.bijectors.Shift(0.1),
          name=constants.EC_RF,
      ),
  )
  ec_om: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.TruncatedNormal(
          0.8, 0.8, 0.1, 10, name=constants.EC_OM
      ),
  )
  ec_orf: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.TransformedDistribution(
          backend.tfd.LogNormal(0.7, 0.4),
          backend.bijectors.Shift(0.1),
          name=constants.EC_ORF,
      ),
  )
  slope_m: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.Deterministic(
          1.0, name=constants.SLOPE_M
      ),
  )
  slope_rf: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.LogNormal(
          0.7, 0.4, name=constants.SLOPE_RF
      ),
  )
  slope_om: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.Deterministic(
          1.0, name=constants.SLOPE_OM
      ),
  )
  slope_orf: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.LogNormal(
          0.7, 0.4, name=constants.SLOPE_ORF
      ),
  )
  sigma: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.HalfNormal(5.0, name=constants.SIGMA),
  )
  roi_m: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.LogNormal(
          0.2, 0.9, name=constants.ROI_M
      ),
  )
  roi_rf: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.LogNormal(
          0.2, 0.9, name=constants.ROI_RF
      ),
  )
  mroi_m: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.LogNormal(
          0.0, 0.5, name=constants.MROI_M
      ),
  )
  mroi_rf: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.LogNormal(
          0.0, 0.5, name=constants.MROI_RF
      ),
  )
  contribution_m: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.Beta(
          1.0, 99.0, name=constants.CONTRIBUTION_M
      ),
  )
  contribution_rf: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.Beta(
          1.0, 99.0, name=constants.CONTRIBUTION_RF
      ),
  )
  contribution_om: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.Beta(
          1.0, 99.0, name=constants.CONTRIBUTION_OM
      ),
  )
  contribution_orf: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.Beta(
          1.0, 99.0, name=constants.CONTRIBUTION_ORF
      ),
  )
  contribution_n: backend.tfd.Distribution = dataclasses.field(
      default_factory=lambda: backend.tfd.TruncatedNormal(
          loc=0.0, scale=0.1, low=-1.0, high=1.0, name=constants.CONTRIBUTION_N
      ),
  )

  def __post_init__(self):
    for param, bounds in _parameter_space_bounds.items():
      prevent_deterministic_prior_at_bounds = (
          _prevent_deterministic_prior_at_bounds[param]
          if param in _prevent_deterministic_prior_at_bounds.keys()
          else (False, False)
      )
      _validate_support(
          param,
          getattr(self, param),
          bounds,
          prevent_deterministic_prior_at_bounds,
      )

  def __setstate__(self, state):
    # Override to support pickling.
    def _unpack_distribution_params(
        params: MutableMapping[str, Any],
    ) -> backend.tfd.Distribution:
      if constants.DISTRIBUTION in params:
        params[constants.DISTRIBUTION] = _unpack_distribution_params(
            params[constants.DISTRIBUTION]
        )
      dist_type = params.pop(constants.DISTRIBUTION_TYPE)
      return dist_type(**params)

    new_state = {}
    for attribute, value in state.items():
      new_state[attribute] = _unpack_distribution_params(value)

    self.__dict__.update(new_state)

  def __getstate__(self):
    # Override to support pickling.
    state = self.__dict__.copy()

    def _pack_distribution_params(
        dist: backend.tfd.Distribution,
    ) -> MutableMapping[str, Any]:
      params = dist.parameters
      params[constants.DISTRIBUTION_TYPE] = type(dist)
      if constants.DISTRIBUTION in params:
        params[constants.DISTRIBUTION] = _pack_distribution_params(
            dist.distribution
        )
      return params

    for attribute, value in state.items():
      state[attribute] = _pack_distribution_params(value)

    return state

  def has_deterministic_param(self, param: backend.tfd.Distribution) -> bool:
    return hasattr(self, param) and isinstance(
        getattr(self, param).distribution, backend.tfd.Deterministic
    )

  def broadcast(
      self,
      n_geos: int,
      n_media_channels: int,
      n_rf_channels: int,
      n_organic_media_channels: int,
      n_organic_rf_channels: int,
      n_controls: int,
      n_non_media_channels: int,
      unique_sigma_for_each_geo: bool,
      n_knots: int,
      is_national: bool,
      set_total_media_contribution_prior: bool,
      kpi: float,
      total_spend: np.ndarray,
  ) -> PriorDistribution:
    """Returns a new `PriorDistribution` with broadcast distribution attributes.

    Args:
      n_geos: Number of geos.
      n_media_channels: Number of media channels used.
      n_rf_channels: Number of reach and frequency channels used.
      n_organic_media_channels: Number of organic media channels used.
      n_organic_rf_channels: Number of organic reach and frequency channels
        used.
      n_controls: Number of controls used.
      n_non_media_channels: Number of non-media channels used.
      unique_sigma_for_each_geo: A boolean indicator whether to use the same
        sigma parameter for all geos. Only used if `n_geos > 1`. For more
        information, see `ModelSpec`.
      n_knots: Number of knots used.
      is_national: A boolean indicator whether the prior distribution will be
        adapted for a national model.
      set_total_media_contribution_prior: A boolean indicator whether the ROI
        priors should be set to achieve a total media constribution prior with
        target mean and variance.
      kpi: Sum of the entire KPI across geos and time. Required if
        `set_total_media_contribution_prior=True`.
      total_spend: Spend per media channel summed across geos and time. Required
        if `set_total_media_contribution_prior=True`.

    Returns:
      A new `PriorDistribution` broadcast from this prior distribution,
      according to the given data dimensionality.

    Raises:
      ValueError: If custom priors are not set for all channels.
    """

    def _validate_media_custom_priors(
        param: backend.tfd.Distribution,
    ) -> None:
      if (
          param.batch_shape.as_list()
          and n_media_channels != param.batch_shape[0]
      ):
        raise ValueError(
            f'Custom priors length ({param.batch_shape[0]}) must match the '
            f' number of media channels ({n_media_channels}), representing a '
            "a custom prior for each channel. If you can't determine a custom "
            'prior, consider using the default prior for that channel.'
        )

    _validate_media_custom_priors(self.roi_m)
    _validate_media_custom_priors(self.mroi_m)
    _validate_media_custom_priors(self.contribution_m)
    _validate_media_custom_priors(self.alpha_m)
    _validate_media_custom_priors(self.ec_m)
    _validate_media_custom_priors(self.slope_m)
    _validate_media_custom_priors(self.eta_m)
    _validate_media_custom_priors(self.beta_m)

    def _validate_organic_media_custom_priors(
        param: backend.tfd.Distribution,
    ) -> None:
      if (
          param.batch_shape.as_list()
          and n_organic_media_channels != param.batch_shape[0]
      ):
        raise ValueError(
            f'Custom priors length ({param.batch_shape[0]}) must match the '
            f' number of organic media channels ({n_organic_media_channels}), '
            "representing a custom prior for each channel. If you can't "
            'determine a custom prior, consider using the default prior for '
            'that channel.'
        )

    _validate_organic_media_custom_priors(self.contribution_om)
    _validate_organic_media_custom_priors(self.alpha_om)
    _validate_organic_media_custom_priors(self.ec_om)
    _validate_organic_media_custom_priors(self.slope_om)
    _validate_organic_media_custom_priors(self.eta_om)
    _validate_organic_media_custom_priors(self.beta_om)

    def _validate_organic_rf_custom_priors(
        param: backend.tfd.Distribution,
    ) -> None:
      if (
          param.batch_shape.as_list()
          and n_organic_rf_channels != param.batch_shape[0]
      ):
        raise ValueError(
            f'Custom priors length ({param.batch_shape[0]}) must match the '
            f'number of organic RF channels ({n_organic_rf_channels}), '
            "representing a custom prior for each channel. If you can't "
            'determine a custom prior, consider using the default prior '
            'for that channel.'
        )

    _validate_organic_rf_custom_priors(self.contribution_orf)
    _validate_organic_rf_custom_priors(self.alpha_orf)
    _validate_organic_rf_custom_priors(self.ec_orf)
    _validate_organic_rf_custom_priors(self.slope_orf)
    _validate_organic_rf_custom_priors(self.eta_orf)
    _validate_organic_rf_custom_priors(self.beta_orf)

    def _validate_rf_custom_priors(
        param: backend.tfd.Distribution,
    ) -> None:
      if param.batch_shape.as_list() and n_rf_channels != param.batch_shape[0]:
        raise ValueError(
            f'Custom priors length ({param.batch_shape[0]}) must match the '
            f'number of RF channels ({n_rf_channels}), representing a custom '
            "prior for each channel. If you can't determine a custom prior, "
            'consider using the default prior for that channel.'
        )

    _validate_rf_custom_priors(self.roi_rf)
    _validate_rf_custom_priors(self.mroi_rf)
    _validate_rf_custom_priors(self.contribution_rf)
    _validate_rf_custom_priors(self.alpha_rf)
    _validate_rf_custom_priors(self.ec_rf)
    _validate_rf_custom_priors(self.slope_rf)
    _validate_rf_custom_priors(self.eta_rf)
    _validate_rf_custom_priors(self.beta_rf)

    def _validate_control_custom_priors(
        param: backend.tfd.Distribution,
    ) -> None:
      if param.batch_shape.as_list() and n_controls != param.batch_shape[0]:
        raise ValueError(
            f'Custom priors length ({param.batch_shape[0]}) must match the '
            f'number of control variables ({n_controls}), representing a '
            "custom prior for each control variable. If you can't determine a "
            'custom prior, consider using the default prior for that variable.'
        )

    _validate_control_custom_priors(self.gamma_c)
    _validate_control_custom_priors(self.xi_c)

    def _validate_non_media_custom_priors(
        param: backend.tfd.Distribution,
    ) -> None:
      if (
          param.batch_shape.as_list()
          and n_non_media_channels != param.batch_shape[0]
      ):
        raise ValueError(
            f'Custom priors length ({param.batch_shape[0]}) must match the '
            f'number of non-media channels ({n_non_media_channels}), '
            "representing a custom prior for each channel. If you can't "
            'determine a custom prior, consider using the default prior for '
            'that channel.'
        )

    _validate_non_media_custom_priors(self.contribution_n)
    _validate_non_media_custom_priors(self.gamma_n)
    _validate_non_media_custom_priors(self.xi_n)

    knot_values = backend.tfd.BatchBroadcast(
        self.knot_values,
        n_knots,
        name=constants.KNOT_VALUES,
    )
    if is_national:
      tau_g_converted = _convert_to_deterministic_0_distribution(
          self.tau_g_excl_baseline
      )
    else:
      tau_g_converted = self.tau_g_excl_baseline
    tau_g_excl_baseline = backend.tfd.BatchBroadcast(
        tau_g_converted, n_geos - 1, name=constants.TAU_G_EXCL_BASELINE
    )
    beta_m = backend.tfd.BatchBroadcast(
        self.beta_m, n_media_channels, name=constants.BETA_M
    )
    beta_rf = backend.tfd.BatchBroadcast(
        self.beta_rf, n_rf_channels, name=constants.BETA_RF
    )
    beta_om = backend.tfd.BatchBroadcast(
        self.beta_om, n_organic_media_channels, name=constants.BETA_OM
    )
    beta_orf = backend.tfd.BatchBroadcast(
        self.beta_orf, n_organic_rf_channels, name=constants.BETA_ORF
    )
    if is_national:
      eta_m_converted = _convert_to_deterministic_0_distribution(self.eta_m)
      eta_rf_converted = _convert_to_deterministic_0_distribution(self.eta_rf)
      eta_om_converted = _convert_to_deterministic_0_distribution(self.eta_om)
      eta_orf_converted = _convert_to_deterministic_0_distribution(self.eta_orf)
    else:
      eta_m_converted = self.eta_m
      eta_rf_converted = self.eta_rf
      eta_om_converted = self.eta_om
      eta_orf_converted = self.eta_orf
    eta_m = backend.tfd.BatchBroadcast(
        eta_m_converted, n_media_channels, name=constants.ETA_M
    )
    eta_rf = backend.tfd.BatchBroadcast(
        eta_rf_converted, n_rf_channels, name=constants.ETA_RF
    )
    eta_om = backend.tfd.BatchBroadcast(
        eta_om_converted,
        n_organic_media_channels,
        name=constants.ETA_OM,
    )
    eta_orf = backend.tfd.BatchBroadcast(
        eta_orf_converted, n_organic_rf_channels, name=constants.ETA_ORF
    )
    gamma_c = backend.tfd.BatchBroadcast(
        self.gamma_c, n_controls, name=constants.GAMMA_C
    )
    if is_national:
      xi_c_converted = _convert_to_deterministic_0_distribution(self.xi_c)
    else:
      xi_c_converted = self.xi_c
    xi_c = backend.tfd.BatchBroadcast(
        xi_c_converted, n_controls, name=constants.XI_C
    )
    gamma_n = backend.tfd.BatchBroadcast(
        self.gamma_n, n_non_media_channels, name=constants.GAMMA_N
    )
    if is_national:
      xi_n_converted = _convert_to_deterministic_0_distribution(self.xi_n)
    else:
      xi_n_converted = self.xi_n
    xi_n = backend.tfd.BatchBroadcast(
        xi_n_converted, n_non_media_channels, name=constants.XI_N
    )
    alpha_m = backend.tfd.BatchBroadcast(
        self.alpha_m, n_media_channels, name=constants.ALPHA_M
    )
    alpha_rf = backend.tfd.BatchBroadcast(
        self.alpha_rf, n_rf_channels, name=constants.ALPHA_RF
    )
    alpha_om = backend.tfd.BatchBroadcast(
        self.alpha_om, n_organic_media_channels, name=constants.ALPHA_OM
    )
    alpha_orf = backend.tfd.BatchBroadcast(
        self.alpha_orf, n_organic_rf_channels, name=constants.ALPHA_ORF
    )
    ec_m = backend.tfd.BatchBroadcast(
        self.ec_m, n_media_channels, name=constants.EC_M
    )
    ec_rf = backend.tfd.BatchBroadcast(
        self.ec_rf, n_rf_channels, name=constants.EC_RF
    )
    ec_om = backend.tfd.BatchBroadcast(
        self.ec_om, n_organic_media_channels, name=constants.EC_OM
    )
    ec_orf = backend.tfd.BatchBroadcast(
        self.ec_orf, n_organic_rf_channels, name=constants.EC_ORF
    )
    if (
        not isinstance(self.slope_m, backend.tfd.Deterministic)
        or (backend.rank(self.slope_m.loc) == 0 and self.slope_m.loc != 1.0)
        or (
            self.slope_m.batch_shape.as_list()
            and any(x != 1.0 for x in self.slope_m.loc)
        )
    ):
      warnings.warn(
          'Changing the prior for `slope_m` may lead to convex Hill curves.'
          ' This may lead to poor MCMC convergence and budget optimization'
          ' may no longer produce a global optimum.'
      )
    slope_m = backend.tfd.BatchBroadcast(
        self.slope_m, n_media_channels, name=constants.SLOPE_M
    )
    slope_rf = backend.tfd.BatchBroadcast(
        self.slope_rf, n_rf_channels, name=constants.SLOPE_RF
    )
    if (
        not isinstance(self.slope_om, backend.tfd.Deterministic)
        or (backend.rank(self.slope_om.loc) == 0 and self.slope_om.loc != 1.0)
        or (
            self.slope_om.batch_shape.as_list()
            and any(x != 1.0 for x in self.slope_om.loc)
        )
    ):
      warnings.warn(
          'Changing the prior for `slope_om` may lead to convex Hill curves.'
          ' This may lead to poor MCMC convergence and budget optimization'
          ' may no longer produce a global optimum.'
      )
    slope_om = backend.tfd.BatchBroadcast(
        self.slope_om, n_organic_media_channels, name=constants.SLOPE_OM
    )
    slope_orf = backend.tfd.BatchBroadcast(
        self.slope_orf, n_organic_rf_channels, name=constants.SLOPE_ORF
    )

    # If `unique_sigma_for_each_geo == False`, then make a scalar batch.
    sigma_shape = n_geos if (n_geos > 1 and unique_sigma_for_each_geo) else []
    sigma = backend.tfd.BatchBroadcast(
        self.sigma, sigma_shape, name=constants.SIGMA
    )

    if set_total_media_contribution_prior:
      roi_m_converted = _get_total_media_contribution_prior(
          kpi, total_spend, constants.ROI_M
      )
      roi_rf_converted = _get_total_media_contribution_prior(
          kpi, total_spend, constants.ROI_RF
      )
    else:
      roi_m_converted = self.roi_m
      roi_rf_converted = self.roi_rf
    roi_m = backend.tfd.BatchBroadcast(
        roi_m_converted, n_media_channels, name=constants.ROI_M
    )
    roi_rf = backend.tfd.BatchBroadcast(
        roi_rf_converted, n_rf_channels, name=constants.ROI_RF
    )

    mroi_m = backend.tfd.BatchBroadcast(
        self.mroi_m, n_media_channels, name=constants.MROI_M
    )
    mroi_rf = backend.tfd.BatchBroadcast(
        self.mroi_rf, n_rf_channels, name=constants.MROI_RF
    )

    contribution_m = backend.tfd.BatchBroadcast(
        self.contribution_m, n_media_channels, name=constants.CONTRIBUTION_M
    )
    contribution_rf = backend.tfd.BatchBroadcast(
        self.contribution_rf, n_rf_channels, name=constants.CONTRIBUTION_RF
    )
    contribution_om = backend.tfd.BatchBroadcast(
        self.contribution_om,
        n_organic_media_channels,
        name=constants.CONTRIBUTION_OM,
    )
    contribution_orf = backend.tfd.BatchBroadcast(
        self.contribution_orf,
        n_organic_rf_channels,
        name=constants.CONTRIBUTION_ORF,
    )
    contribution_n = backend.tfd.BatchBroadcast(
        self.contribution_n, n_non_media_channels, name=constants.CONTRIBUTION_N
    )

    return PriorDistribution(
        knot_values=knot_values,
        tau_g_excl_baseline=tau_g_excl_baseline,
        beta_m=beta_m,
        beta_rf=beta_rf,
        beta_om=beta_om,
        beta_orf=beta_orf,
        eta_m=eta_m,
        eta_rf=eta_rf,
        eta_om=eta_om,
        eta_orf=eta_orf,
        gamma_c=gamma_c,
        gamma_n=gamma_n,
        xi_c=xi_c,
        xi_n=xi_n,
        alpha_m=alpha_m,
        alpha_rf=alpha_rf,
        alpha_om=alpha_om,
        alpha_orf=alpha_orf,
        ec_m=ec_m,
        ec_rf=ec_rf,
        ec_om=ec_om,
        ec_orf=ec_orf,
        slope_m=slope_m,
        slope_rf=slope_rf,
        slope_om=slope_om,
        slope_orf=slope_orf,
        sigma=sigma,
        roi_m=roi_m,
        roi_rf=roi_rf,
        mroi_m=mroi_m,
        mroi_rf=mroi_rf,
        contribution_m=contribution_m,
        contribution_rf=contribution_rf,
        contribution_om=contribution_om,
        contribution_orf=contribution_orf,
        contribution_n=contribution_n,
    )


class IndependentMultivariateDistribution(backend.tfd.Distribution):
  """Container for a joint distribution created from independent distributions.

  This class is useful when one wants to define a joint distribution for a
  Meridian prior, where the elements are not necessarily from the same
  distribution family. For example, to define a distribution where
  one element is Uniform and the second is triangular:

  ```python
  distributions = [
      tfp.distributions.Uniform(0.0, 1.0),
      tfp.distributions.Triangular(0.0, 1.0, 0.5)
      ]
  distribution = IndependentMultivariateDistribution(distributions)
  ```

  It is also possible to define a distribution where multiple elements come
  from the same distribution family. For example, to define a distribution where
  the three elements are LogNormal(0.2, 0.9), LogNormal(0, 0.5) and
  Gamma(2, 2):

  ```python
  distributions = [
      tfp.distributions.LogNormal([0.2, 0.0], [0.9, 0.5]),
      tfp.distributions.Gamma(2.0, 2.0)
      ]
  distribution = IndependentMultivariateDistribution(distributions)
  ```

  This class cannot contain instances of `tfd.Deterministic`.
  """

  def __init__(
      self,
      distributions: Sequence[backend.tfd.Distribution],
      validate_args: bool = False,
      allow_nan_stats: bool = True,
      name: str | None = None,
  ):
    """Initializes a batch of independent distributions from different families.

    Args:
      distributions: List of `tfd.Distribution` from which to construct a
        multivariate distribution. The distributions must have scalar or one
        dimensional batch shapes; the resulting batch shape will be the sum of
        the underlying batch shapes.
      validate_args: Python `bool`. When `True` distribution parameters are
        checked for validity despite possibly degrading runtime performance.
        When `False` invalid inputs may silently render incorrect outputs.
        Default value is `False`.
      allow_nan_stats: Python `bool`. When `True`, statistics (e.g., mean, mode,
        variance) use the value "`NaN`" to indicate the result is undefined.
        When `False`, an exception is raised if one or more of the statistic's
        batch members are undefined. Default value is `True`.
      name: Python `str` name prefixed to Ops created by this class. Default
        value is 'IndependentMultivariate' followed by the names of the
        underlying distributions.

    Raises:
        ValueError: If one or more distributions are instances of
        `tfd.Deterministic` or dtypes differ between the
        distributions.
    """
    parameters = dict(locals())

    self._verify_distributions(distributions)

    self._distributions = [
        dist
        if not dist.is_scalar_batch()
        else backend.tfd.BatchBroadcast(dist, (1,))
        for dist in distributions
    ]

    self._distribution_batch_shapes = self._get_distribution_batch_shapes()
    self._distribution_batch_shape_tensors = backend.concatenate(
        [dist.batch_shape_tensor() for dist in self._distributions],
        axis=0,
    )

    dtype = self._verify_dtypes()

    name = name or '-'.join(
        [constants.INDEPENDENT_MULTIVARIATE] + [d.name for d in distributions]
    )

    super().__init__(
        dtype=dtype,
        reparameterization_type=backend.tfd.NOT_REPARAMETERIZED,
        validate_args=validate_args,
        allow_nan_stats=allow_nan_stats,
        parameters=parameters,
        name=name,
    )

  def _verify_distributions(
      self, distributions: Sequence[backend.tfd.Distribution]
  ):
    """Check for deterministic distributions and raise an error if found."""

    if any(
        isinstance(dist, backend.tfd.Deterministic) for dist in distributions
    ):
      raise ValueError(
          f'{self.__class__.__name__} cannot contain `Deterministic` '
          'distributions. To implement a nearly deterministic element of this '
          'distribution, we recommend using `backend.tfd.Uniform` with a '
          'small range. For example to define a distribution that is nearly '
          '`Deterministic(1.0)`, use '
          '`tfp.distribution.Uniform(1.0 - 1e-9, 1.0 + 1e-9)`'
      )

  def _verify_dtypes(self) -> str:
    dtypes = [dist.dtype for dist in self._distributions]
    if len(set(dtypes)) != 1:
      raise ValueError(
          f'All distributions must have the same dtype. Found: {dtypes}.'
      )

    return backend.result_type(*dtypes)

  def _event_shape(self):
    return backend.TensorShape([])

  def _batch_shape_tensor(self):
    distribution_batch_shape_tensors = backend.concatenate(
        [dist.batch_shape_tensor() for dist in self._distributions],
        axis=0,
    )
    return backend.reduce_sum(distribution_batch_shape_tensors, keepdims=True)

  def _batch_shape(self):
    return backend.TensorShape(sum(self._distribution_batch_shapes))

  def _sample_n(self, n, seed=None):
    return backend.concatenate(
        [dist.sample(n, seed) for dist in self._distributions], axis=-1
    )

  def _quantile(self, value):
    value = self._broadcast_value(value)
    split_value = backend.split(value, self._distribution_batch_shapes, axis=-1)
    quantiles = [
        dist.quantile(sv) for dist, sv in zip(self._distributions, split_value)
    ]

    return backend.concatenate(quantiles, axis=-1)

  def _log_prob(self, value):
    value = self._broadcast_value(value)
    split_value = backend.split(value, self._distribution_batch_shapes, axis=-1)
    log_probs = [
        dist.log_prob(sv) for dist, sv in zip(self._distributions, split_value)
    ]

    return backend.concatenate(log_probs, axis=-1)

  def _log_cdf(self, value):
    value = self._broadcast_value(value)
    split_value = backend.split(value, self._distribution_batch_shapes, axis=-1)

    log_cdfs = [
        dist.log_cdf(sv) for dist, sv in zip(self._distributions, split_value)
    ]

    return backend.concatenate(log_cdfs, axis=-1)

  def _mean(self):
    return backend.concatenate(
        [dist.mean() for dist in self._distributions], axis=0
    )

  def _variance(self):
    return backend.concatenate(
        [dist.variance() for dist in self._distributions], axis=0
    )

  def _default_event_space_bijector(self):
    """Mapping from R^n to the event space of the wrapped distributions.

    This is the blockwise concatenation of the underlying bijectors.

    Returns:
      A `tfp.bijectors.Blockwise` object that concatenates the underlying
      bijectors.
    """
    bijectors = [
        d.experimental_default_event_space_bijector()
        for d in self._distributions
    ]

    return backend.bijectors.Blockwise(
        bijectors,
        block_sizes=self._distribution_batch_shapes,
    )

  def _broadcast_value(self, value: backend.Tensor) -> backend.Tensor:
    value = backend.to_tensor(value)
    broadcast_shape = backend.broadcast_dynamic_shape(
        value.shape, self.batch_shape_tensor()
    )
    return backend.broadcast_to(value, broadcast_shape)

  def _get_distribution_batch_shapes(self) -> Sequence[int]:
    """Sequence of batch shapes of underlying distributions."""

    batch_shapes = []

    for dist in self._distributions:
      try:
        (dist_batch_shape,) = dist.batch_shape
      except ValueError as exc:
        raise ValueError(
            'All distributions must be 0- or 1-dimensional.'
            f' Found {len(dist.batch_shape)}-dimensional distribution:'
            f' {dist.batch_shape}.'
        ) from exc
      else:
        batch_shapes.append(dist_batch_shape)

    return batch_shapes


def distributions_are_equal(
    a: backend.tfd.Distribution, b: backend.tfd.Distribution
) -> bool:
  """Determine if two distributions are equal."""
  if type(a) != type(b):  # pylint: disable=unidiomatic-typecheck
    return False

  a_params = a.parameters.copy()
  b_params = b.parameters.copy()

  if constants.DISTRIBUTION in a_params and constants.DISTRIBUTION in b_params:
    if not distributions_are_equal(
        a_params[constants.DISTRIBUTION], b_params[constants.DISTRIBUTION]
    ):
      return False
    del a_params[constants.DISTRIBUTION]
    del b_params[constants.DISTRIBUTION]

  if constants.DISTRIBUTION in a_params or constants.DISTRIBUTION in b_params:
    return False

  if a_params.keys() != b_params.keys():
    return False

  for key in a_params.keys():
    if isinstance(
        a_params[key], (backend.Tensor, np.ndarray, float, int)
    ) and isinstance(b_params[key], (backend.Tensor, np.ndarray, float, int)):
      if not backend.allclose(a_params[key], b_params[key]):
        return False
    else:
      if a_params[key] != b_params[key]:
        return False

  return True


def lognormal_dist_from_mean_std(
    mean: float | Sequence[float], std: float | Sequence[float]
) -> backend.tfd.LogNormal:
  """Define a lognormal distribution from its mean and standard deviation.

  This function parameterizes lognormal distributions by their mean and
  standard deviation.

  Args:
    mean: A float or array-like object defining the distribution mean. Must be
      positive.
    std: A float or array-like object defining the distribution standard
      deviation. Must be non-negative.

  Returns:
    A `backend.tfd.LogNormal` object with the input mean and standard deviation.
  """

  mean = np.asarray(mean)
  std = np.asarray(std)

  mu = np.log(mean) - 0.5 * np.log((std / mean) ** 2 + 1)
  sigma = np.sqrt(np.log((std / mean) ** 2 + 1))

  return backend.tfd.LogNormal(mu, sigma)


def lognormal_dist_from_range(
    low: float | Sequence[float],
    high: float | Sequence[float],
    mass_percent: float | Sequence[float] = 0.95,
) -> backend.tfd.LogNormal:
  """Define a LogNormal distribution from a specified range.

  This function parameterizes lognormal distributions by the bounds of a range,
  so that the specified probability mass falls within the bounds defined by
  `low` and `high`. The probability mass is symmetric about the median. For
  example, to define a lognormal distribution with a 95% probability mass of
  (1, 10), use:

  ```python
  lognormal = lognormal_dist_from_range(1.0, 10.0, mass_percent=0.95)
  ```

  Args:
    low: Float or array-like denoting the lower bound of the range. Values must
      be non-negative.
    high: Float or array-like denoting the upper bound of range. Values must be
      non-negative.
    mass_percent: Float or array-like denoting the probability mass. Values must
      be between 0 and 1 (exclusive). Default: 0.95.

  Returns:
    A `backend.tfd.LogNormal` object with the input percentage mass falling
      within the given range.
  """
  low = np.asarray(low)
  high = np.asarray(high)
  mass_percent = np.asarray(mass_percent)

  if not ((0.0 < low).all() and (low < high).all()):  # pytype: disable=attribute-error
    raise ValueError("'low' and 'high' values must be non-negative and satisfy "
                     "high > low.")

  if not ((0.0 < mass_percent).all() and (mass_percent < 1.0).all()):  # pytype: disable=attribute-error
    raise ValueError(
        "'mass_percent' values must be between 0 and 1, exclusive."
        )

  normal = backend.tfd.Normal(0, 1)
  mass_lower = 0.5 - (mass_percent / 2)
  mass_upper = 0.5 + (mass_percent / 2)

  sigma = np.log(high / low) / (
      normal.quantile(mass_upper) - normal.quantile(mass_lower)
  )
  mu = np.log(high) - normal.quantile(mass_upper) * sigma

  return backend.tfd.LogNormal(mu, sigma)


def _convert_to_deterministic_0_distribution(
    distribution: backend.tfd.Distribution,
) -> backend.tfd.Distribution:
  """Converts the given distribution to a `Deterministic(0)` one.

  Args:
    distribution: `tfp.distributions.Distribution` object to be converted to
      `Deterministic(0)` distribution.

  Returns:
    `tfp.distribution.Deterministic(0, distribution.name)`

  Raises:
    Warning: If the argument distribution is not a `Deterministic(0)`
    distribution.
  """
  if (
      not isinstance(distribution, backend.tfd.Deterministic)
      or distribution.loc != 0
  ):
    warnings.warn(
        'Hierarchical distribution parameters must be deterministically zero'
        f' for national models. {distribution.name} has been automatically set'
        ' to Deterministic(0).'
    )
    return backend.tfd.Deterministic(loc=0, name=distribution.name)
  else:
    return distribution


def _get_total_media_contribution_prior(
    kpi: float,
    total_spend: np.ndarray,
    name: str,
    p_mean: float = constants.P_MEAN,
    p_sd: float = constants.P_SD,
) -> backend.tfd.Distribution:
  """Determines ROI priors based on total media contribution.

  Args:
    kpi: Sum of the entire KPI across geos and time.
    total_spend: Spend per media channel summed across geos and time.
    name: Name of the distribution.
    p_mean: Prior mean proportion of KPI incremental due to all media. Default
      value is `0.4`.
    p_sd: Prior standard deviation proportion of KPI incremental to all media.
      Default value is `0.2`.

  Returns:
    A new `Distribution` based on total media contribution.
  """
  roi_mean = p_mean * kpi / np.sum(total_spend)
  roi_sd = p_sd * kpi / np.sqrt(np.sum(np.power(total_spend, 2)))
  lognormal_sigma = backend.cast(
      np.sqrt(np.log(roi_sd**2 / roi_mean**2 + 1)), dtype=backend.float32
  )
  lognormal_mu = backend.cast(
      np.log(roi_mean * np.exp(-(lognormal_sigma**2) / 2)),
      dtype=backend.float32,
  )
  return backend.tfd.LogNormal(lognormal_mu, lognormal_sigma, name=name)


def _validate_support(
    parameter_name: str,
    tfp_dist: backend.tfp.distributions.Distribution,
    bounds: tuple[float, float],
    prevent_deterministic_prior_at_bounds: tuple[bool, bool],
) -> None:
  """Validates that distribution support is within the parameter bounds.

  Args:
    parameter_name: Name of the parameter.
    tfp_dist: The TFP distribution to validate.
    bounds: Tuple containing the min and max values of the parameteter space.
    prevent_deterministic_prior_at_bounds: Tuple of two booleans indicating
      whether a deterministic prior is allowed at the lower and upper bounds,
      respectively.

  Raises:
    ValueError: If the distribution support is not within the parameter bounds.
  """
  # Note that `tfp.distributions.BatchBroadcast` objects have a `distribution`
  # attribute that points to a `tfp.distributions.Distribution` object.
  if isinstance(tfp_dist, backend.tfd.BatchBroadcast):
    tfp_dist = tfp_dist.distribution
  # Note that `tfp.distributions.Deterministic` does not have a `quantile`
  # method implemented, so the min and max values must be extracted from the
  # `loc` attribute instead.
  if isinstance(tfp_dist, backend.tfd.Deterministic):
    support_min_vals = tfp_dist.loc
    support_max_vals = tfp_dist.loc
    for i in (0, 1):
      if prevent_deterministic_prior_at_bounds[i] and np.any(
          tfp_dist.loc == bounds[i]
      ):
        raise ValueError(
            f'{parameter_name} was assigned a point mass (deterministic) prior'
            f' at {bounds[i]}, which is not allowed.'
        )
  elif isinstance(tfp_dist, backend.tfd.TruncatedNormal):
    # TruncatedNormal quantile method is not reliable, particularly when the
    # `low` or `high` value falls into extreme percentile of the untruncated
    # distribution. Note that
    # `TruncatedNormal.experimental_default_event_space_bijector()([-inf, inf])`
    # returns the correct support range, so this method could be used if the
    # `quantile` method is found to be unreliable for other distributions.
    support_min_vals = tfp_dist.low
    support_max_vals = tfp_dist.high
  else:
    try:
      support_min_vals = tfp_dist.quantile(0)
      support_max_vals = tfp_dist.quantile(1)
    except (AttributeError, NotImplementedError):
      warnings.warn(
          f'The prior distribution for {parameter_name} does not have a'
          ' `quantile` method implemented, so the support range validation'
          f' was skipped. Confirm that your prior for {parameter_name} is'
          ' appropriate.'
      )
      return
  if np.any(support_min_vals < bounds[0]):
    raise ValueError(
        f'{parameter_name} was assigned a prior distribution that allows values'
        f' less than the parameter minimum {bounds[0]}.'
    )
  if np.any(support_max_vals > bounds[1]):
    raise ValueError(
        f'{parameter_name} was assigned a prior distribution that allows values'
        f' greater than the parameter maximum {bounds[1]}.'
    )


# Dictionary of parameters that have a limited parameters space. The tuple
# contains the lower and upper bounds, respectively.
_parameter_space_bounds = {
    'eta_m': (0, np.inf),
    'eta_rf': (0, np.inf),
    'eta_om': (0, np.inf),
    'eta_orf': (0, np.inf),
    'xi_c': (0, np.inf),
    'xi_n': (0, np.inf),
    'alpha_m': (0, 1),
    'alpha_rf': (0, 1),
    'alpha_om': (0, 1),
    'alpha_orf': (0, 1),
    'ec_m': (0, np.inf),
    'ec_rf': (0, np.inf),
    'ec_om': (0, np.inf),
    'ec_orf': (0, np.inf),
    'slope_m': (0, np.inf),
    'slope_rf': (0, np.inf),
    'slope_om': (0, np.inf),
    'slope_orf': (0, np.inf),
    'sigma': (0, np.inf),
}

# Dictionary of parameters that do not allow a deterministic prior at one or
# more of the parameter space bounds. The boolean tuple indicates whether a
# deterministic prior is allowed at the lower bound or upper bound,
# respectively, where `True` means "not allowed". This check is specifically for
# point mass at finite paramteter space bounds, since point mass at infinity is
# generally problematic for all parameters. Note that `sigma` should generally
# not have point mass at zero, but this is not checked here because unit tests
# require the ability to simulate data with `sigma` set to zero.
_prevent_deterministic_prior_at_bounds = {
    'alpha_m': (False, True),
    'alpha_rf': (False, True),
    'alpha_om': (False, True),
    'alpha_orf': (False, True),
    'ec_m': (True, False),
    'ec_rf': (True, False),
    'ec_om': (True, False),
    'ec_orf': (True, False),
    'slope_m': (True, False),
    'slope_rf': (True, False),
    'slope_om': (True, False),
    'slope_orf': (True, False),
}



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Auxiliary functions for knots calculations."""

from collections.abc import Collection, Sequence
import copy
import dataclasses
import math
import pprint
from typing import Any

from meridian import constants
from meridian.data import input_data
import numpy as np
from scipy import interpolate
from statsmodels.regression import linear_model


__all__ = [
    'KnotInfo',
    'get_knot_info',
    'l1_distance_weights',
]


def _find_left_knot_indices(
    *,
    times: np.ndarray,
    knot_locations: np.ndarray,
) -> Sequence[int]:
  """Return the index of the left neighboring knot for each time point.

  Args:
    times: Times `0, 1, 2,..., (n_times-1)`.
    knot_locations: The location of knots within `0, 1, 2,..., (n_times-1)`.

  Returns:
    A list of indices of the left neighboring knot for each time point. The
    length of the list is equal to the length of `times`.
    - If a time point is at or before the first knot, the index is 0.
    - If a time point is at or after the last knot, the index is `n_knots - 1`.
    - Otherwise, it's the index of the knot just to the left.
  """
  n_knots = len(knot_locations)
  # Find indices such that knot_locations[i-1] < times <= knot_locations[i]
  insert_indices = np.searchsorted(knot_locations, times, side='right')
  left_knot_indices = insert_indices - 1

  # Handle edge cases for times before the first knot
  left_knot_indices[times < knot_locations[0]] = 0
  # Handle edge cases for times at or after the last knot
  left_knot_indices[times >= knot_locations[-1]] = n_knots - 1

  return left_knot_indices


def l1_distance_weights(
    n_times: int, knot_locations: np.ndarray[int, np.dtype[int]]
) -> np.ndarray:
  """Computes weights at knots for every time period.

  The two neighboring knots inform the weight estimate of a particular time
  period. The amount each of the two neighboring knots inform the weight at a
  time period depends on how close (L1 distance) they are to the time period. If
  a time point coincides with a knot location, then 100% weight is given to that
  knot. If a time point lies outside the range of knots, then 100% weight is
  given to the nearest endpoint knot.

  This function computes an `(n_knots, n_times)` array of weights that are used
  to model trend and seasonality. For a given time, the array contains two
  non-zero weights. The weights are inversely proportional to the L1 distance
  from the given time to the neighboring knots. The two weights are normalized
  such that they sum to 1.

  Args:
    n_times: The number of time points.
    knot_locations: The location of knots within `0, 1, 2,..., (n_times-1)`.

  Returns:
    A weight array with dimensions `(n_knots, n_times)` with values summing up
    to 1 for each time period when summing over knots.
  """
  if knot_locations.ndim != 1:
    raise ValueError('`knot_locations` must be one-dimensional.')
  if not np.all(knot_locations == np.sort(knot_locations)):
    raise ValueError('`knot_locations` must be sorted.')
  if len(knot_locations) <= 1:
    raise ValueError('Number of knots must be greater than 1.')
  if len(knot_locations) != len(np.unique(knot_locations)):
    raise ValueError('`knot_locations` must be unique.')
  if np.any(knot_locations < 0):
    raise ValueError('knot_locations must be positive.')
  if np.any(knot_locations >= n_times):
    raise ValueError('knot_locations must be less than `n_times`.')

  times = np.arange(n_times)
  time_minus_knot = abs(knot_locations[:, np.newaxis] - times[np.newaxis, :])

  w = np.zeros(time_minus_knot.shape, dtype=np.float32)
  left_knot_indices = _find_left_knot_indices(
      times=times, knot_locations=knot_locations
  )

  for t in times:
    left_idx = left_knot_indices[t]
    current_time = times[t]

    if current_time in knot_locations:
      # If time is exactly at a knot, give all weight to that knot.
      knot_idx = np.where(knot_locations == current_time)[0][0]
      w[knot_idx, t] = 1.0
    elif current_time < knot_locations[0] or current_time > knot_locations[-1]:
      # Outside the knot range, assign full weight to the closest endpoint knot.
      w[left_idx, t] = 1.0
    else:
      # Time is between left_idx and left_idx + 1.
      left_dist = time_minus_knot[left_idx, t]
      right_dist = time_minus_knot[left_idx + 1, t]
      total_dist = left_dist + right_dist

      # Assign weight inversely proportional to distance.
      # The closer knot gets more weight.
      w[left_idx, t] = right_dist / total_dist
      w[left_idx + 1, t] = left_dist / total_dist

  return w


def _get_equally_spaced_knot_locations(n_times, n_knots):
  """Equally spaced knot locations starting at the endpoints."""
  return np.linspace(0, n_times - 1, n_knots, dtype=int)


@dataclasses.dataclass(frozen=True)
class KnotInfo:
  """Contains the number of knots, knot locations, and weights.

  Attributes:
    n_knots: The number of knots
    knot_locations: The location of knots
    weights: The weights used to multiply with the knot values to get time-
      varying coefficients.
  """

  n_knots: int
  knot_locations: np.ndarray[int, np.dtype[int]]
  weights: np.ndarray[int, np.dtype[float]]


def get_knot_info(
    n_times: int,
    knots: int | Collection[int] | None,
    enable_aks: bool = False,
    data: input_data.InputData | None = None,
    is_national: bool = False,
) -> KnotInfo:
  """Returns the number of knots, knot locations, and weights.

  Args:
    n_times: The number of time periods in the data.
    knots: An optional integer or a collection of integers indicating the knots
      used to estimate time effects. When `knots` is a collection of integers,
      the knot locations are provided by that collection. Zero corresponds to a
      knot at the first time period, one corresponds to a knot at the second
      time, ..., and `(n_times - 1)` corresponds to a knot at the last time
      period. When `knots` is an integer, then there are knots with locations
      equally spaced across the time periods (including knots at zero and
      `(n_times - 1)`. When `knots` is `1`, there is a single common regression
      coefficient used for all time periods. If `knots` is `None`, then the
      numbers of knots used is equal to the number of time periods. This is
      equivalent to each time period having its own regression coefficient.
    enable_aks: A boolean indicating whether to use the Automatic Knot Selection
      algorithm to select optimal number of knots for running the model instead
      of the default 1 for national and n_times for non-national models.
    data: An Optional InputData object used by the Automatic Knot Selection
      algorithm to calculate optimal number of knots from the provided Input
      Data.
    is_national: A boolean indicator whether to adapt the knot information for a
      national model.

  Returns:
    A KnotInfo that contains the number of knots, the location of knots, and the
    weights used to multiply with the knot values to get time-varying
    coefficients.
  """
  if enable_aks:
    if data is None:
      raise ValueError(
          'If enable_aks is true then input data must be provided.'
      )
    else:
      aks = AKS(data)
      knots = aks.automatic_knot_selection().knots
      n_knots = len(knots)
      knot_locations = knots
  elif isinstance(knots, int):
    if knots < 1:
      raise ValueError('If knots is an integer, it must be at least 1.')
    elif knots > n_times:
      raise ValueError(
          f'The number of knots ({knots}) cannot be greater than the number of'
          f' time periods in the kpi ({n_times}).'
      )
    elif is_national and knots == n_times:
      raise ValueError(
          f'Number of knots ({knots}) must be less than number of time periods'
          f' ({n_times}) in a nationally aggregated model.'
      )
    n_knots = knots
    knot_locations = _get_equally_spaced_knot_locations(n_times, n_knots)
  elif isinstance(knots, Collection) and knots:
    if any(k < 0 for k in knots):
      raise ValueError('Knots must be all non-negative.')
    if any(k >= n_times for k in knots):
      raise ValueError(
          'Knots must all be less than the number of time periods.'
      )
    n_knots = len(knots)
    # np.unique also sorts
    knot_locations = np.unique(knots)
  elif isinstance(knots, Collection):
    raise ValueError('Knots cannot be empty.')
  else:
    # knots is None
    n_knots = 1 if is_national else n_times
    knot_locations = _get_equally_spaced_knot_locations(n_times, n_knots)

  if n_knots == 1:
    weights = np.ones((1, n_times), dtype=np.float32)
  else:
    weights = l1_distance_weights(n_times, knot_locations)

  return KnotInfo(n_knots, knot_locations, weights)


@dataclasses.dataclass(frozen=True)
class AKSResult:
  knots: np.ndarray[int, np.dtype[int]]
  model: linear_model.OLS


class AKS:
  """Class for automatically selecting knots in Meridian Core Library."""

  _BASE_PENALTY = np.logspace(-1, 2, 100)
  _DEGREE = 1

  def __init__(self, data: input_data.InputData):
    self._data = data

  def automatic_knot_selection(
      self,
      base_penalty: np.ndarray | None = None,
      min_internal_knots: int = 1,
      max_internal_knots: int | None = None,
  ) -> AKSResult:
    """Calculates the optimal number of knots for Meridian model using Automatic knot selection with A-spline.

    Args:
      base_penalty: A vector of positive penalty values. The adaptive spline
        regression is performed for every value of penalty.
      min_internal_knots: The minimum number of internal knots. Defaults to 1.
      max_internal_knots: The maximum number of internal knots. If None, this
        value is calculated as the number of initial knots minus the total count
        of all treatment and control variables. Otherwise, the user-provided
        value will be used.

    Returns:
      Selected knots and the corresponding B-spline model.
    """
    if base_penalty is None:
      base_penalty = self._BASE_PENALTY
    n_times = len(self._data.time)
    n_geos = len(self._data.geo)

    y_tensor = self._data.scaled_centered_kpi
    y = np.reshape(y_tensor, (n_geos * n_times,))
    x = np.reshape(
        np.repeat([range(n_times)], n_geos, axis=0), (n_geos * n_times,)
    )

    knots = self._calculate_initial_knots(x)
    max_internal_knots = self._calculate_and_validate_max_internal_knots(
        knots, min_internal_knots, max_internal_knots
    )
    geo_scaling_factor = 1 / np.sqrt(len(self._data.geo))
    penalty = geo_scaling_factor * base_penalty

    aspline = self.aspline(x=x, y=y, knots=knots, penalty=penalty)
    # Ensure defined knot range covers at least one of the available knot sets.
    available_knots_lengths = np.unique(
        np.fromiter(
            (len(x) for x in aspline[constants.KNOTS_SELECTED]), dtype=int
        )
    ).tolist()
    if not any(
        min_internal_knots <= k <= max_internal_knots
        for k in available_knots_lengths
    ):
      raise ValueError(
          f'The range [{min_internal_knots}, {max_internal_knots}] does not'
          ' contain any of the available knot lengths:'
          f' {pprint.pformat(available_knots_lengths)}'
      )

    n_knots = np.array([len(x) for x in aspline[constants.KNOTS_SELECTED]])
    feasible_idx = np.where(
        (n_knots >= min_internal_knots) & (n_knots <= max_internal_knots)
    )[0]
    information_criterion = aspline[constants.AIC][feasible_idx]
    knots_sel = [aspline[constants.KNOTS_SELECTED][i] for i in feasible_idx]
    model = [aspline[constants.MODEL][i] for i in feasible_idx]
    opt_idx = max(
        np.where(information_criterion == min(information_criterion))[0]
    )

    return AKSResult(knots_sel[opt_idx], model[opt_idx])

  def _get_bspline_matrix(self, x, knots):
    """Replaces patsy.highlevel.dmatrix('bs(...)', ...)"""
    # Pad knots at boundaries to match patsy's 'bs()' behavior
    t = np.concatenate((
        [x.min()] * (self._DEGREE + 1),
        np.sort(knots),
        [x.max()] * (self._DEGREE + 1),
    ))
    # Create design matrix (standard numpy array)
    return interpolate.BSpline.design_matrix(x, t, self._DEGREE).toarray()

  def _calculate_initial_knots(
      self,
      x: np.ndarray,
  ) -> np.ndarray:
    """Calculates initial knots based on unique x values.

    Args:
      x: A flattened array of indexed time coordinates, repeated n_geos times.
        e.g. [0, 1, 2, 3, ..., 0, 1, 2, 3, ...].

    Returns:
      A tuple containing:
        - The calculated knots.
    """
    x_vals_unique = np.unique(x)
    min_x_data, max_x_data = x_vals_unique.min(), x_vals_unique.max()
    knots = x_vals_unique[
        (x_vals_unique > min_x_data) & (x_vals_unique < max_x_data)
    ]
    knots = np.sort(np.unique(knots))
    # Drop one knot from the set of all knots because the algorithm requires one
    # fewer degree of freedom than the total number of knots to function.
    # Dropping the final knot is a natural and practical choice because it
    # often has minimal impact on the overall model fit.
    return knots[:-1]

  def _calculate_and_validate_max_internal_knots(
      self,
      knots: np.ndarray,
      min_internal_knots: int = 1,
      max_internal_knots: int | None = None,
  ) -> int:
    """Calculates the max internal knots, and validates the range.

    Args:
      knots: Initial knots to calculate max internal knots from.
      min_internal_knots: The minimum number of internal knots.
      max_internal_knots: The maximum number of internal knots. If None, this
        value will be calculated, otherwise will use user provided value.

    Returns:
      The maximum number of internal knots.
    """
    n_features = sum(
        len(feature) if feature is not None else 0
        for feature in [
            self._data.media_channel,
            self._data.rf_channel,
            self._data.organic_media_channel,
            self._data.organic_rf_channel,
            self._data.non_media_channel,
            self._data.control_variable,
        ]
    )

    if max_internal_knots is None:
      max_internal_knots = len(knots) - n_features
    if min_internal_knots > len(knots):
      raise ValueError(
          'The minimum number of internal knots cannot be greater than the'
          ' total number of initial knots.'
      )
    if max_internal_knots < min_internal_knots:
      raise ValueError(
          'The maximum number of internal knots cannot be less than the minimum'
          ' number of internal knots.'
      )
    return max_internal_knots

  def aspline(
      self,
      x: np.ndarray,
      y: np.ndarray,
      knots: np.ndarray,
      penalty: np.ndarray,
      max_iterations: int = 1000,
      epsilon: float = 1e-5,
      tol: float = 1e-6,
  ) -> dict[str, Any]:
    """Fits B-splines with automatic knot selection.

    Args:
      x: A flattened array of indexed time coordinates, repeated n_geos times.
        e.g. [0, 1, 2, 3, ..., 0, 1, 2, 3, ...].
      y: The flattened array of KPI values that have been population-scaled and
        mean-centered by geo.
      knots: Internal knots used for spline regression.
      penalty: A vector of positive penalty values. The adaptive spline
        regression is performed for every value of penalty.
      max_iterations: Maximum number of iterations in the main loop.
      epsilon: Value of the constant in the adaptive ridge procedure (see
        Frommlet, F., Nuel, G. (2016) An Adaptive Ridge Procedure for L0
        Regularization.)
      tol: The tolerance chosen to diagnose convergence of the adaptive ridge
        procedure.

    Returns:
      A dictionary of the following items:
        selection_coefs: A list of selection coefficients for every value of
        penalty.
        knots_selected: A list of selected knots for every value of penalty.
        model: A list of fitted models for every value of penalty.
        regression_coefs: A list of estimated regression coefficients for every
        value of penalty.
        selected_matrix: A matrix of selected knots for every value of penalty.
        aic: A list of AIC values for every value of penalty.
        bic: A list of BIC values for every value of penalty.
        ebic: A list of EBIC values for every value of penalty.
    """
    if x.ndim != 1 or y.ndim != 1:
      raise ValueError(
          'Provided x and y args for aspline must both be 1 dimensional!'
      )

    xmat = self._get_bspline_matrix(x, knots)
    nrow = xmat.shape[0]
    ncol = xmat.shape[1]

    xx = xmat.T.dot(xmat)
    xy = xmat.T.dot(y)
    xx_rot = np.concat(
        [
            self._mat2rot(xx + (1e-20 * np.identity(ncol))),
            np.zeros(ncol)[:, np.newaxis],
        ],
        axis=1,
    )
    sigma0sq = linear_model.OLS(y, xmat).fit().mse_resid ** 2
    model, x_sel, knots_sel, sel_ls, par_ls, aic, bic, ebic, dim, loglik = (
        [None] * len(penalty) for _ in range(10)
    )
    old_sel, w = [np.ones(ncol - self._DEGREE - 1) for _ in range(2)]
    par = np.ones(ncol)
    index_penalty = 0
    for _ in range(max_iterations):
      par = self._wridge_solver(
          xx_rot, xy, self._DEGREE, penalty[index_penalty], w, old_par=par
      )
      par_diff = np.diff(par, n=self._DEGREE + 1)

      w = 1 / (par_diff**2 + epsilon**2)
      sel = w * par_diff**2
      converge = max(abs(old_sel - sel)) < tol
      if converge:
        sel_ls[index_penalty] = sel
        knots_sel[index_penalty] = knots[sel > 0.99]

        design_mat = self._get_bspline_matrix(x, knots_sel[index_penalty])
        x_sel[index_penalty] = design_mat
        bs_model = linear_model.OLS(y, x_sel[index_penalty]).fit()
        model[index_penalty] = bs_model
        coefs = np.zeros(ncol, dtype=np.float32)
        idx = np.concat([sel > 0.99, np.repeat(True, self._DEGREE + 1)])
        coefs[idx] = bs_model.params
        par_ls[index_penalty] = coefs

        loglik[index_penalty] = sum(bs_model.resid**2 / sigma0sq) / 2
        dim[index_penalty] = len(knots_sel[index_penalty]) + self._DEGREE + 1
        aic[index_penalty] = 2 * dim[index_penalty] + 2 * loglik[index_penalty]
        bic[index_penalty] = (
            np.log(nrow) * dim[index_penalty] + 2 * loglik[index_penalty]
        )
        ebic[index_penalty] = bic[index_penalty] + 2 * np.log(
            np.float32(math.comb(ncol, design_mat.shape[1]))
        )
        index_penalty = index_penalty + 1
      if index_penalty > len(penalty) - 1:
        break
      old_sel = sel

    sel_mat = np.round(np.stack(sel_ls, axis=-1), 1)
    return {
        constants.SELECTION_COEFS: sel_ls,
        constants.KNOTS_SELECTED: knots_sel,
        constants.MODEL: model,
        constants.REGRESSION_COEFS: par_ls,
        constants.SELECTED_MATRIX: sel_mat,
        constants.AIC: np.array(aic),
        constants.BIC: np.array(bic),
        constants.EBIC: np.array(ebic),
    }

  def _mat2rot(self, band_mat: np.ndarray) -> np.ndarray:
    """Rotates a symmetric band matrix to get the rotated matrix associated.

    Each column of the rotated matrix corresponds to a diagonal. The first
    column is the main diagonal, the second one is the upper-diagonal and so on.
    Artificial 0s are placed at the end of each column if necessary.

    Args:
      band_mat: The band square matrix to be rotated.

    Returns:
      The rotated matrix of band_mat.
    """
    p = band_mat.shape[1]
    l = 0
    for i in range(p):
      lprime = np.where(band_mat[i, :] != 0)[0]
      l = np.maximum(l, lprime[len(lprime) - 1] - i)

    rot_mat = np.zeros([p, l + 1])
    rot_mat[:, 0] = np.diag(band_mat)
    if l > 0:
      for j in range(l):
        rot_mat[:, j + 1] = np.concat([
            np.diag(band_mat[range(p - j - 1), :][:, range(j + 1, p)]),
            np.zeros(j + 1),
        ])
    return rot_mat

  def _band_weight(self, w: np.ndarray, diff: int) -> np.ndarray:
    """Creates the penalty matrix for A-Spline.

    Args:
      w: Vector of weights.
      diff: Order of the differences to be applied to the parameters. Must be a
        strictly positive integer.

    Returns:
      Weighted penalty matrix D'diag(w)D, where
      D = diff(diag(len(w) + diff), differences = diff)}. Only the non-null
      superdiagonals of the weight matrix are returned, each column
      corresponding to a diagonal.
    """
    ws = len(w)
    rows = ws + diff
    cols = diff + 1

    # Compute the entries of the difference matrix
    binom = np.zeros(cols, dtype=np.int32)
    for i in range(cols):
      binom[i] = math.comb(diff, i) * (-1) ** i

    # Compute the limit indices
    ind_mat = np.zeros([rows, 2], dtype=np.int32)
    for ind in range(rows):
      ind_mat[ind, 0] = 0 if ind - diff < 0 else ind - diff
      ind_mat[ind, 1] = ind if ind < ws - 1 else ws - 1

    # Main loop
    result = np.zeros([rows, cols])
    for j in range(cols):
      for i in range(rows - j):
        temp = 0.0
        for k in range(ind_mat[i + j, 0], ind_mat[i, 1] + 1):
          temp += binom[i - k] * binom[i + j - k] * w[k]
        result[i, j] = temp

    return result

  def _ldl(self, rot_mat: np.ndarray) -> np.ndarray:
    """Solves the Fast LDL decomposition of symmetric band matrix of length k.

    Args:
      rot_mat: Rotated row-wised matrix of dimensions n*k, with first column
        corresponding to the diagonal, the second to the first super-diagonal
        and so on.

    Returns:
      Solution of the LDL decomposition.
    """
    n = rot_mat.shape[0]
    m = rot_mat.shape[1] - 1
    rot_mat_new = copy.deepcopy(rot_mat)
    for i in range(1, n + 1):
      j0 = np.maximum(1, i - m)
      for j in range(j0, i + 1):
        for k in range(j0, j):
          rot_mat_new[j - 1, i - j] -= (
              rot_mat_new[k - 1, i - k]
              * rot_mat_new[k - 1, j - k]
              * rot_mat_new[k - 1, 0]
          )
        if i > j:
          rot_mat_new[j - 1, i - j] /= rot_mat_new[j - 1, 0]

    return rot_mat_new

  def _bandsolve_kernel(
      self, rot_mat: np.ndarray, rhs_mat: np.ndarray
  ) -> np.ndarray:
    """Solves the symmetric bandlinear system Ax = b.

    This is the kernel function that solves the system, where A is the rotated
    form of the band matrix and b is the right hand side.

    Args:
      rot_mat: Band square matrix in the rotated form. It's the visual rotation
        by 90 degrees of the matrix, where subdiagonal are discarded.
      rhs_mat: right hand side of the equation. Can be either a vector or a
        matrix. If not supplied, the function return the inverse of rot_mat.

    Returns:
      Solution of the linear problem.
    """
    rot_mat_ldl = self._ldl(rot_mat)
    x = copy.deepcopy(rhs_mat)
    n = rot_mat.shape[0]
    k = rot_mat_ldl.shape[1] - 1
    l = rhs_mat.shape[1]

    for l in range(l):
      # solve b=inv(L)b
      for i in range(2, n + 1):
        jmax = np.minimum(i - 1, k)
        for j in range(1, jmax + 1):
          x[i - 1, l] -= rot_mat_ldl[i - j - 1, j] * x[i - j - 1, l]

      # solve b=b/D
      for i in range(n):
        x[i, l] /= rot_mat_ldl[i, 0]

      # solve b=inv(t(L))b=inv(L*D*t(L))b
      for i in range(n - 1, 0, -1):
        jmax = np.minimum(n - i, k)
        for j in range(1, jmax + 1):
          x[i - 1, l] -= rot_mat_ldl[i - 1, j] * x[i + j - 1, l]

    return x

  def _bandsolve(self, rot_mat: np.ndarray, rhs_mat: np.ndarray) -> np.ndarray:
    """Solves the symmetric bandlinear system Ax = b.

    Here A is the rotated form of the band matrix and b is the right hand side.

    Args:
      rot_mat: Band square matrix in the rotated form. It's the visual rotation
        by 90 degrees of the matrix, where subdiagonal are discarded.
      rhs_mat: right hand side of the equation. Can be either a vector or a
        matrix. If not supplied, the function return the inverse of rot_mat.

    Returns:
      Solution of the linear problem.
    """

    nrow = rot_mat.shape[0]
    ncol = rot_mat.shape[1]
    if (nrow == ncol) & (rot_mat[nrow - 1, ncol - 1] != 0):
      raise ValueError('rot_mat should be a rotated matrix!')
    if rot_mat[nrow - 1, 1] != 0:
      raise ValueError('rot_mat should be a rotated matrix!')
    if len(rhs_mat) != nrow:
      raise ValueError('Dimension problem!')

    return self._bandsolve_kernel(rot_mat, rhs_mat[:, np.newaxis])

  def _wridge_solver(
      self,
      xx_rot: np.ndarray,
      xy: np.ndarray,
      degree: int,
      penalty: float,
      w: np.ndarray,
      old_par: np.ndarray,
      max_iterations: int = 1000,
      tol: float = 1e-8,
  ) -> np.ndarray | None:
    """Fits B-Splines with weighted penalization over differences of parameters.

    Args:
      xx_rot: The matrix X'X where X is the design matrix. This argument is
        given in the form of a band matrix, i.e., successive columns represent
        superdiagonals.
      xy: The vector of currently estimated points X'y, where y is the
        y-coordinate of the data.
      degree: The degree of the B-splines.
      penalty: Positive penalty constant.
      w: Vector of weights. The case w = np.ones(xx_rot.shape[0] - degree - 1)
        corresponds to fitting P-splines with difference order degree + 1. See
        Eilers, P., Marx, B. (1996) Flexible smoothing with B-splines and
        penalties.
      old_par: The previous parameter vector.
      max_iterations: Maximum number of Newton-Raphson iterations to be
        computed.
      tol: The tolerance chosen to diagnose convergence of the adaptive ridge
        procedure.

    Returns:
      The estimated parameter of the spline regression.
    """

    def _hessian_solver(par, xx_rot, xy, penalty, w, diff):
      """Inverts the hessian and multiplies it by the score.

      Args:
        par: The parameter vector.
        xx_rot: The matrix X'X where X is the design matrix. This argument is
          given in the form of a rotated band matrix, i.e., successive columns
          represent superdiagonals.
        xy: The vector of currently estimated points X'y, where y is the
          y-coordinate of the data.
        penalty: Positive penalty constant.
        w: Vector of weights.
        diff: The order of the differences of the parameter. Equals degree + 1
          in adaptive spline regression.

      Returns:
        The solution of the linear system: (X'X + penalty*D'WD)^{-1} X'y - par
      """
      if xx_rot.shape[1] != diff + 1:
        raise ValueError('Error: xx_rot must have diff + 1 columns')
      return (
          self._bandsolve(xx_rot + penalty * self._band_weight(w, diff), xy)[
              :, 0
          ]
          - par
      )

    par = None
    for _ in range(max_iterations):
      par = old_par + _hessian_solver(
          par=old_par,
          xx_rot=xx_rot,
          xy=xy,
          penalty=penalty,
          w=w,
          diff=degree + 1,
      )
      index = old_par != 0
      rel_error = max(abs(par - old_par)[index] / abs(old_par)[index])
      if rel_error < tol:
        break
      old_par = par

    return par



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Constants shared across the Meridian serde library."""

# Constants for hyperparameters protobuf structure
BASELINE_GEO_ONEOF = 'baseline_geo_oneof'
BASELINE_GEO_INT = 'baseline_geo_int'
BASELINE_GEO_STRING = 'baseline_geo_string'
CONTROL_POPULATION_SCALING_ID = 'control_population_scaling_id'
HOLDOUT_ID = 'holdout_id'
NON_MEDIA_POPULATION_SCALING_ID = 'non_media_population_scaling_id'
ADSTOCK_DECAY_SPEC = 'adstock_decay_spec'
GLOBAL_ADSTOCK_DECAY = 'global_adstock_decay'
ADSTOCK_DECAY_BY_CHANNEL = 'adstock_decay_by_channel'
DEFAULT_DECAY = 'geometric'

# Constants for marketing data protobuf structure
GEO_INFO = 'geo_info'
METADATA = 'metadata'
REACH_FREQUENCY = 'reach_frequency'

# Constants for distribution protobuf structure
DISTRIBUTION_TYPE = 'distribution_type'
BATCH_BROADCAST_DISTRIBUTION = 'batch_broadcast'
DETERMINISTIC_DISTRIBUTION = 'deterministic'
HALF_NORMAL_DISTRIBUTION = 'half_normal'
LOG_NORMAL_DISTRIBUTION = 'log_normal'
NORMAL_DISTRIBUTION = 'normal'
TRANSFORMED_DISTRIBUTION = 'transformed'
TRUNCATED_NORMAL_DISTRIBUTION = 'truncated_normal'
UNIFORM_DISTRIBUTION = 'uniform'
BETA_DISTRIBUTION = 'beta'
BIJECTOR_TYPE = 'bijector_type'
SHIFT_BIJECTOR = 'shift'
SCALE_BIJECTOR = 'scale'
RECIPROCAL_BIJECTOR = 'reciprocal'
